# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
This file implements a DC-OPF for time series
That means that solves the OPF problem for a complete time series at once
"""
import numpy as np
from itertools import product

from GridCal.Engine.basic_structures import ZonalGrouping
from GridCal.Engine.Simulations.OPF.opf_templates import OpfTimeSeries, LpProblem, LpVariable, Logger
from GridCal.Engine.basic_structures import MIPSolvers
from GridCal.Engine.Core.time_series_opf_data import OpfTimeCircuit
import GridCal.ThirdParty.pulp as pl
from GridCal.Engine.Devices.enumerations import TransformerControlType, ConverterControlType, HvdcControlType, GenerationNtcFormulation


def get_objective_function(Pg, Pb, LSlack, FSlack1, FSlack2, FCSlack1, FCSlack2,
                           flow_from_a1_to_a2, sum_gen_area_1, sum_gen_area_2,
                           cost_g, cost_b, cost_l, cost_br):
    """
    Add the objective function to the problem
    :param Pg: generator LpVars (ng, nt)
    :param Pb: batteries LpVars (nb, nt)
    :param LSlack: Load slack LpVars (nl, nt)
    :param FSlack1: Branch overload slack1 (m, nt)
    :param FSlack2: Branch overload slack2 (m, nt)
    :param FCSlack1: Branch contingency overload slack1 (m, nt)
    :param FCSlack2: Branch contingency overload slack2 (m, nt)
    :param cost_g: Cost of the generators (ng, nt)
    :param cost_b: Cost of the batteries (nb, nt)
    :param cost_l: Cost of the loss of load (nl, nt)
    :param cost_br: Cost of the overload (m, nt)
    :return: Nothing, just assign the objective function
    """

    f_obj = pl.lpSum(cost_g * Pg)

    f_obj += pl.lpSum(cost_b * Pb)

    f_obj += pl.lpSum(cost_l * LSlack)

    f_obj += pl.lpSum(cost_br * (FSlack1 + FSlack2))

    f_obj += cost_br * pl.lpSum(FCSlack1) + cost_br * pl.lpSum(FCSlack2)

    if flow_from_a1_to_a2 is not None:
        f_obj -= pl.lpSum(flow_from_a1_to_a2)  # maximize

    if sum_gen_area_1 is not None:
        f_obj -= pl.lpSum(sum_gen_area_1)  # maximize
        f_obj += pl.lpSum(sum_gen_area_2)  # minimize

    return f_obj


def set_fix_generation(problem, Pg, P_profile, enabled_for_dispatch):
    """
    Set the generation fixed at the non dispatchable generators
    :param problem: LP problem instance
    :param Pg: Array of generation variables
    :param P_profile: Array of fixed generation values
    :param enabled_for_dispatch: array of "enables" for dispatching generators
    :return: Nothing
    """

    idx = np.where(enabled_for_dispatch == False)[0]

    pl.lpAddRestrictions2(problem=problem,
                          lhs=Pg[idx, :],
                          rhs=P_profile[idx, :],
                          # Fmax + FSlack2
                          name='fixed_generation',
                          op='=')


def get_power_injections(C_bus_gen, Pg, C_bus_bat, Pb, C_bus_load, LSlack, Pl):
    """
    Create the power injections per bus
    :param C_bus_gen: Bus-Generators sparse connectivity matrix (n, ng)
    :param Pg: generator LpVars (ng, nt)
    :param C_bus_bat: Bus-Batteries sparse connectivity matrix (n, nb)
    :param Pb: Batteries LpVars (nb, nt)
    :param C_bus_load: Bus-Load sparse connectivity matrix (n, nl)
    :param LSlack: Load slack LpVars (nl, nt)
    :param Pl: Load values (nl, nt)
    :return: Power injection at the buses (n, nt)
    """

    P = pl.lpDot(C_bus_gen, Pg)

    P += pl.lpDot(C_bus_bat, Pb)

    P -= pl.lpDot(C_bus_load, Pl - LSlack)

    return P


def formulate_dc_nodal_power_balance(numerical_circuit: OpfTimeCircuit, problem: LpProblem, theta, P, start_, end_):
    """
    Add the nodal power balance
    :param numerical_circuit: NumericalCircuit instance
    :param problem: LpProblem instance
    :param theta: Voltage angles LpVars (n, nt)
    :param P: Power injection at the buses LpVars (n, nt)
    :return: Nothing, the restrictions are added to the problem (nbus, nt)
    """

    # do the topological computation
    calc_inputs = numerical_circuit.split_into_islands(ignore_single_node_islands=True)

    # generate the time indices to simulate
    if end_ == -1:
        end_ = len(numerical_circuit.time_array)

    # For every island, run the time series
    nodal_restrictions = np.empty((numerical_circuit.nbus, end_ - start_), dtype=object)
    for i, calc_inpt in enumerate(calc_inputs):

        # find the original indices
        bus_original_idx = np.array(calc_inpt.original_bus_idx)

        # re-pack the variables for the island and time interval
        P_island = P[bus_original_idx, :]  # the sizes already reflect the correct time span
        theta_island = theta[bus_original_idx, :]  # the sizes already reflect the correct time span
        B_island = calc_inpt.Ybus.imag

        pqpv = calc_inpt.pqpv
        vd = calc_inpt.vd

        # Add nodal power balance for the non slack nodes
        idx = bus_original_idx[pqpv]
        nodal_restrictions[idx] = pl.lpAddRestrictions2(problem=problem,
                                                        lhs=pl.lpDot(B_island[np.ix_(pqpv, pqpv)], theta_island[pqpv, :]),
                                                        rhs=P_island[pqpv, :],
                                                        name='Nodal_power_balance_pqpv_is' + str(i),
                                                        op='=')

        # Add nodal power balance for the slack nodes
        idx = bus_original_idx[vd]
        nodal_restrictions[idx] = pl.lpAddRestrictions2(problem=problem,
                                                        lhs=pl.lpDot(B_island[vd, :], theta_island),
                                                        rhs=P_island[vd, :],
                                                        name='Nodal_power_balance_vd_is' + str(i),
                                                        op='=')

        # slack angles equal to zero
        pl.lpAddRestrictions2(problem=problem,
                              lhs=theta_island[vd, :],
                              rhs=np.zeros_like(theta_island[vd, :]),
                              name='Theta_vd_zero_is' + str(i),
                              op='=')

    return nodal_restrictions


def add_branch_loading_restriction(problem: LpProblem,
                                   nc: OpfTimeCircuit,
                                   theta, F, T,
                                   ratings, ratings_slack_from, ratings_slack_to,
                                   monitored, active):
    """
    Add the branch loading restrictions
    :param problem: LpProblem instance
    :param nc: OpfTimeCircuit instance
    :param theta: array of LpVariables with the bus angles (n, nt)
    :param F: Array with the "from" branch indices (m)
    :param T: Array with the "to" branch indices (m)
    :param ratings: Array of branch ratings (m, nt)
    :param ratings_slack_from: Array of branch loading slack variables in the from-to sense
    :param ratings_slack_to: Array of branch loading slack variables in the to-from sense
    :param monitored: Array with the monitoring status (m, nt)
    :param active: Array with the active status (m, nt)
    :return: Pbr_f: Array of power flowing through the branch (m, nt)
             tau: Array of tap angles in the phase shifters (m, nt)
             Pinj_tau: Array of the nodal power injections due to the phase shift (n, nt)
    """

    nbr, nt = ratings_slack_to.shape

    # from-to branch power restriction
    Pbr_f = np.zeros((nbr, nt), dtype=object)
    tau = np.zeros((nbr, nt), dtype=object)
    Pinj_tau = np.zeros((nc.nbus, nt), dtype=object)

    for m, t in product(range(nbr), range(nt)):
        if active[m, t]:

            # compute the branch susceptance
            if nc.branch_data.branch_dc[m]:
                bk = -1.0 / nc.branch_data.R[m]
            else:
                bk = -1.0 / nc.branch_data.X[m]

            # compute the flow
            if nc.branch_data.control_mode[m] == TransformerControlType.Pt:
                # is a phase shifter device (like phase shifter transformer or VSC with P control)
                tau[m, t] = LpVariable('Tau_{0}_{1}'.format(m, t), nc.branch_data.theta_min[m], nc.branch_data.theta_max[m])
                Pbr_f[m, t] = bk * (theta[F[m], t] - theta[T[m], t] + tau[m, t])

                # power injected and subtracted due to the phase shift
                Pinj_tau[F[m], t] = -bk * tau[m, t]
                Pinj_tau[T[m], t] = bk * tau[m, t]

            else:
                # regular branch
                tau[m, t] = 0.0
                Pbr_f[m, t] = bk * (theta[F[m], t] - theta[T[m], t])

            if monitored[m]:
                problem.add(Pbr_f[m, t] <= ratings[m, t] + ratings_slack_from[m, t], 'upper_rate_{0}_{1}'.format(m, t))
                problem.add(-ratings[m, t] - ratings_slack_to[m, t] <= Pbr_f[m, t], 'lower_rate_{0}_{1}'.format(m, t))
        else:
            Pbr_f[m, t] = 0

    return Pbr_f, tau, Pinj_tau


def formulate_contingency(problem: LpProblem, numerical_circuit: OpfTimeCircuit, flow_f, ratings, LODF, monitor,
                          lodf_tolerance):
    """

    :param problem:
    :param numerical_circuit:
    :param flow_f:
    :param LODF:
    :param monitor:
    :return:
    """
    nbr, nt = ratings.shape

    # get the indices of the branches marked for contingency
    con_br_idx = numerical_circuit.branch_data.get_contingency_enabled_indices()

    # formulate contingency flows
    # this is done in a separated loop because all te flow variables must exist beforehand
    flow_lst = list()
    indices = list()  # (t, m, contingency_m)
    overload1_lst = list()
    overload2_lst = list()

    for t, m in product(range(nt), range(nbr)):  # for every branch

        if monitor[m]:  # the monitor variable is pre-computed in the previous loop
            _f = numerical_circuit.branch_data.F[m]
            _t = numerical_circuit.branch_data.T[m]

            for ic, c in enumerate(con_br_idx):  # for every contingency

                if m != c and abs(LODF[m, c]) >= lodf_tolerance:

                    # compute the N-1 flow
                    contingency_flow = flow_f[m, t] + LODF[m, c] * flow_f[c, t]

                    # rating restriction in the sense from-to
                    overload1 = LpVariable("n-1_overload1_{0}_{1}_{2}".format(t, m, c), 0, 99999)
                    problem.add(contingency_flow <= (ratings[m, t] + overload1),
                                "n-1_ft_up_rating_{0}_{1}_{2}".format(t, m, c))

                    # rating restriction in the sense to-from
                    overload2 = LpVariable("n-1_overload2_{0}_{1}_{2}".format(t, m, c), 0, 99999)
                    problem.add((-ratings[m, t] - overload2) <= contingency_flow,
                                "n-1_tf_down_rating_{0}_{1}_{2}".format(t, m, c))

                    # store the variables
                    flow_lst.append(contingency_flow)
                    overload1_lst.append(overload1)
                    overload2_lst.append(overload2)
                    indices.append((t, m, c))

    return flow_lst, overload1_lst, overload2_lst, indices


def add_battery_discharge_restriction(problem: LpProblem, SoC0, Capacity, Efficiency, Pb, E, dt):
    """
    Add the batteries capacity restrictions
    :param problem: LpProblem instance
    :param SoC0: State of Charge at 0 (nb)
    :param SoCmax: Maximum State of Charge (nb)
    :param SoCmin: Minimum State of Charge (nb)
    :param Capacity: Capacities of the batteries (nb) in MWh/MW base
    :param Efficiency: Roundtrip efficiency
    :param Pb: Batteries injection power LpVars (nb, nt)
    :param E: Batteries Energy state LpVars (nb, nt)
    :param dt: time increments in hours (nt-1)
    :return: Nothing, the restrictions are added to the problem
    """

    # set the initial state of charge
    pl.lpAddRestrictions2(problem=problem,
                          lhs=E[:, 0],
                          rhs=SoC0 * Capacity,
                          name='initial_soc',
                          op='=')

    # compute the inverse of he efficiency because pulp does not divide by floats
    eff_inv = 1 / Efficiency

    # set the Energy values for t=1:nt
    for i in range(len(dt) - 1):

        t = i + 1

        # set the energy value Et = E(t-1) + dt * Pb / eff
        pl.lpAddRestrictions2(problem=problem,
                              lhs=E[:, t],
                              rhs=E[:, t - 1] - dt[i] * Pb[:, t] * eff_inv,
                              name='initial_soc_t' + str(t) + '_',
                              op='=')


def formulate_hvdc_flow(problem: LpProblem, angles, Pinj, rates, active, Pset,
                        control_mode, dispatchable, angle_droop, F, T, Sbase,
                        logger: Logger = Logger(), inf=999999):
    """

    :param problem:
    :param angles:
    :param Pinj:
    :param rates:
    :param active:
    :param Pset:
    :param control_mode:
    :param dispatchable:
    :param angle_droop:
    :param F:
    :param T:
    :param logger:
    :param inf:
    :param Sbase:
    :return:
    """
    nhvdc, nt = rates.shape

    flow_f = np.zeros((nhvdc, nt), dtype=object)

    for t, i in product(range(nt), range(nhvdc)):

        if active[i, t]:

            _f = F[i]
            _t = T[i]

            if control_mode[i] == HvdcControlType.type_0_free:

                if rates[i, t] <= 0:
                    logger.add_error('Rate = 0', 'HVDC:{0} t:{1}'.format(i, t), rates[i, t])

                # formulate the hvdc flow as an AC line equivalent
                flow_f[i, t] = LpVariable('flow_hvdc1_{0}_{1}'.format(i, t), -rates[i, t], rates[i, t])
                P0 = Pset[i, t] / Sbase
                problem.add(flow_f[i, t] == P0 + angle_droop[i, t] * (angles[_f, t] - angles[_t, t]),
                            "hvdc_flow_set_{0}_{1}".format(i, t))

                # add the injections matching the flow
                Pinj[_f, t] -= flow_f[i, t]
                Pinj[_t, t] += flow_f[i, t]

                # rating restriction in the sense from-to: eq.17
                # overload1[i, t] = LpVariable('overload_hvdc1_{0}_{1}'.format(i, t), 0, inf)
                # problem.add(flow_f[i, t] <= (rates[i, t] + overload1[i, t]), "hvdc_ft_rating_{0}_{1}".format(i, t))

                # rating restriction in the sense to-from: eq.18
                # overload2[i, t] = LpVariable('overload_hvdc2_{0}_{1}'.format(i, t), 0, inf)
                # problem.add((-rates[i, t] - overload2[i, t]) <= flow_f[i, t], "hvdc_tf_rating_{0}_{1}".format(i, t))

            elif control_mode[i] == HvdcControlType.type_1_Pset and not dispatchable[i]:
                # simple injections model: The power is set by the user
                P0 = Pset[i, t] / Sbase
                flow_f[i, t] = P0  # + hvdc_control1[i, t] - hvdc_control2[i, t]
                Pinj[_f, t] -= flow_f[i, t]
                Pinj[_t, t] += flow_f[i, t]

            elif control_mode[i] == HvdcControlType.type_1_Pset and dispatchable[i]:
                # simple injections model, the power is a variable and it is optimized
                P0 = LpVariable('hvdc_pf_{0}_{1}'.format(i, t), -rates[i, t], rates[i, t])
                flow_f[i, t] = P0
                Pinj[_f, t] -= flow_f[i, t]
                Pinj[_t, t] += flow_f[i, t]

    return flow_f


def formulate_inter_area_flow(numerical_circuit: OpfTimeCircuit,
                              buses_areas_1, buses_areas_2,
                              flow_f, hvdc_flow_f):
    """
    Formulate the flow that goes through the links from the area 1 (from) to the area 2 (to)
    :param numerical_circuit: OpfTimeCircuit instance
    :param buses_areas_1: array of bus indices that compose the area 1 (from)
    :param buses_areas_2: array of bus indices that compose the area 2 (to)
    :param flow_f: array of branch flows
    :param hvdc_flow_f: array of HVDC links flows
    :return: vector per time step of the sum of flows 1->2 with the correct sign as a PuLP equation
    """
    nhvdc, nt = hvdc_flow_f.shape

    inter_area_branches = numerical_circuit.get_inter_areas_branches(buses_areas_1=buses_areas_1,
                                                                     buses_areas_2=buses_areas_2)
    inter_area_hvdc = numerical_circuit.get_inter_areas_hvdc(buses_areas_1=buses_areas_1,
                                                             buses_areas_2=buses_areas_2)

    flows_ft = np.zeros((len(inter_area_branches), nt), dtype=object)
    flows_hvdc_ft = np.zeros((len(inter_area_hvdc), nt), dtype=object)
    flow_from_a1_to_a2 = np.zeros(nt, dtype=object)

    for t in range(nt):

        for i, (k, sign) in enumerate(inter_area_branches):
            flows_ft[i, t] = sign * flow_f[k, t]

        for i, (k, sign) in enumerate(inter_area_hvdc):
            flows_hvdc_ft[i, t] = sign * hvdc_flow_f[k, t]

        flow_from_a1_to_a2[t] = pl.lpSum(flows_ft) + pl.lpSum(flows_hvdc_ft)

    return flow_from_a1_to_a2


def formulate_area_generation_summations(numerical_circuit: OpfTimeCircuit, buses_areas_1, buses_areas_2, Pg):
    """
    Compute the summation of the generation in the area 1 and area 2
    :param numerical_circuit: OpfTimeCircuit instance
    :param buses_areas_1: array of bus indices that compose the area 1 (from)
    :param buses_areas_2: array of bus indices that compose the area 2 (to)
    :param Pg: Array of generator variables
    :return: Summation of generation in the area 1, summation of the generation in the are 2
    """

    _, nt = Pg.shape

    gens_in_a1, gens_in_a2, gens_out = numerical_circuit.get_generators_per_areas(buses_in_a1=buses_areas_1,
                                                                                  buses_in_a2=buses_areas_2)

    bat_in_a1, bat_in_a2, bat_out = numerical_circuit.get_batteries_per_areas(buses_in_a1=buses_areas_1,
                                                                              buses_in_a2=buses_areas_2)

    sum_a1 = np.zeros(nt, dtype=object)
    sum_a2 = np.zeros(nt, dtype=object)

    for t in range(nt):

        for bus_idx, gen_idx in gens_in_a1:
            sum_a1[t] += Pg[gen_idx, t]

        for bus_idx, gen_idx in gens_in_a2:
            sum_a2[t] += Pg[gen_idx, t]

        for bus_idx, gen_idx in bat_in_a1:
            sum_a1[t] += Pg[gen_idx, t]

        for bus_idx, gen_idx in bat_in_a2:
            sum_a2[t] += Pg[gen_idx, t]

    return sum_a1, sum_a2


class OpfDcTimeSeries(OpfTimeSeries):

    def __init__(self, numerical_circuit: OpfTimeCircuit, start_idx, end_idx, solver_type: MIPSolvers = MIPSolvers.CBC,
                 zonal_grouping: ZonalGrouping = ZonalGrouping.NoGrouping,
                 skip_generation_limits=False, consider_contingencies=False, LODF=None, lodf_tolerance=0.001,
                 maximize_inter_area_flow=False,
                 buses_areas_1=None, buses_areas_2=None):
        """
        DC time series linear optimal power flow
        :param numerical_circuit: NumericalCircuit instance
        :param start_idx: start index of the time series
        :param end_idx: end index of the time series
        :param solver_type: MIP solver_type to use
        :param zonal_grouping:
        :param skip_generation_limits:
        :param consider_contingencies:
        :param LODF:
        :param lodf_tolerance:
        :param maximize_inter_area_flow:
        :param buses_areas_1:
        :param buses_areas_2:
        """
        OpfTimeSeries.__init__(self, numerical_circuit=numerical_circuit,
                               start_idx=start_idx, end_idx=end_idx,
                               solver_type=solver_type, skip_formulation=True)

        self.zonal_grouping = zonal_grouping
        self.skip_generation_limits = skip_generation_limits
        self.consider_contingencies = consider_contingencies
        self.LODF = LODF
        self.lodf_tolerance = lodf_tolerance

        self.maximize_inter_area_flow = maximize_inter_area_flow
        self.buses_areas_1: List[int] = buses_areas_1
        self.buses_areas_2: List[int] = buses_areas_2

    def formulate(self, batteries_energy_0=None):
        """
        Formulate the DC OPF time series in the non-sequential fashion (all to the solver_type at once)
        :param batteries_energy_0: initial energy state of the batteries (if none, the default is taken)
        :return: PuLP Problem instance
        """

        # general indices
        n = self.numerical_circuit.nbus
        m = self.numerical_circuit.nbr
        ng = self.numerical_circuit.ngen
        nb = self.numerical_circuit.nbatt
        nl = self.numerical_circuit.nload
        nt = self.end_idx - self.start_idx
        a = self.start_idx
        b = self.end_idx
        Sbase = self.numerical_circuit.Sbase

        # battery
        Capacity = self.numerical_circuit.battery_enom / Sbase
        minSoC = self.numerical_circuit.battery_min_soc
        maxSoC = self.numerical_circuit.battery_max_soc
        if batteries_energy_0 is None:
            SoC0 = self.numerical_circuit.battery_soc_0
        else:
            SoC0 = (batteries_energy_0 / Sbase) / Capacity
        Pb_max = self.numerical_circuit.battery_pmax / Sbase
        Pb_min = self.numerical_circuit.battery_pmin / Sbase
        Efficiency = (self.numerical_circuit.battery_discharge_efficiency + self.numerical_circuit.battery_charge_efficiency) / 2.0
        cost_b = self.numerical_circuit.battery_cost[:, a:b]

        # generator
        if self.skip_generation_limits:
            Pg_max = np.zeros(self.numerical_circuit.ngen) * 99999999.0
            Pg_min = np.zeros(self.numerical_circuit.ngen) * -99999999.0
        else:
            Pg_max = self.numerical_circuit.generator_pmax / Sbase
            Pg_min = self.numerical_circuit.generator_pmin / Sbase

        P_profile = self.numerical_circuit.generator_p[:, a:b] / Sbase
        cost_g = self.numerical_circuit.generator_cost[:, a:b]
        enabled_for_dispatch = self.numerical_circuit.generator_dispatchable

        # load
        Pl = (self.numerical_circuit.load_active[:, a:b] * self.numerical_circuit.load_s.real[:, a:b]) / Sbase
        cost_l = self.numerical_circuit.load_cost[:, a:b]

        # branch
        branch_ratings = self.numerical_circuit.branch_rates[:, a:b] / Sbase
        br_active = self.numerical_circuit.branch_data.branch_active[:, a:b]
        F = self.numerical_circuit.F
        T = self.numerical_circuit.T
        cost_br = self.numerical_circuit.branch_cost[:, a:b]

        # Compute time delta in hours
        dt = np.zeros(nt)  # here nt = end_idx - start_idx
        for t in range(1, nt + 1):
            if a + t < nt:
                dt[t - 1] = (self.numerical_circuit.time_array[a + t] - self.numerical_circuit.time_array[a + t - 1]).seconds / 3600
            else:
                dt[t - 1] = 1.0

        # create LP variables
        Pg = pl.lpMakeVars(name='Pg', shape=(ng, nt), lower=Pg_min, upper=Pg_max)
        Pb = pl.lpMakeVars(name='Pb', shape=(nb, nt), lower=Pb_min, upper=Pb_max)
        E = pl.lpMakeVars(name='E', shape=(nb, nt), lower=Capacity * minSoC, upper=Capacity * maxSoC)
        load_slack = pl.lpMakeVars(name='LSlack', shape=(nl, nt), lower=0, upper=None)
        theta = pl.lpMakeVars(name='theta', shape=(n, nt),
                              lower=self.numerical_circuit.bus_data.angle_min,
                              upper=self.numerical_circuit.bus_data.angle_max)
        branch_rating_slack1 = pl.lpMakeVars(name='FSlack1', shape=(m, nt), lower=0, upper=None)
        branch_rating_slack2 = pl.lpMakeVars(name='FSlack2', shape=(m, nt), lower=0, upper=None)

        # declare problem ----------------------------------------------------------------------------------------------
        self.problem = LpProblem(name='DC_OPF_Time_Series')

        # set the fixed generation values ------------------------------------------------------------------------------
        set_fix_generation(problem=self.problem,
                           Pg=Pg,
                           P_profile=P_profile,
                           enabled_for_dispatch=enabled_for_dispatch)

        # compute the power injections ---------------------------------------------------------------------------------
        P = get_power_injections(C_bus_gen=self.numerical_circuit.generator_data.C_bus_gen,
                                 Pg=Pg,
                                 C_bus_bat=self.numerical_circuit.battery_data.C_bus_batt,
                                 Pb=Pb,
                                 C_bus_load=self.numerical_circuit.load_data.C_bus_load,
                                 LSlack=load_slack,
                                 Pl=Pl)

        # formulate the simple HVDC models -----------------------------------------------------------------------------
        hvdc_flow_f = formulate_hvdc_flow(problem=self.problem,
                                          angles=theta,
                                          Pinj=P,
                                          rates=self.numerical_circuit.hvdc_data.rate[:, a:b] / Sbase,
                                          active=self.numerical_circuit.hvdc_data.active[:, a:b],
                                          Pset=self.numerical_circuit.hvdc_data.Pset[:, a:b],
                                          control_mode=self.numerical_circuit.hvdc_data.control_mode,
                                          dispatchable=self.numerical_circuit.hvdc_data.dispatchable,
                                          angle_droop=self.numerical_circuit.hvdc_data.get_angle_droop_in_pu_rad(Sbase),
                                          F=self.numerical_circuit.hvdc_data.get_bus_indices_f(),
                                          T=self.numerical_circuit.hvdc_data.get_bus_indices_t(),
                                          Sbase=Sbase,
                                          logger=self.logger,
                                          inf=999999)

        # add branch restrictions --------------------------------------------------------------------------------------
        if self.zonal_grouping == ZonalGrouping.NoGrouping:
            flow_f, tau, Pinj_tau = add_branch_loading_restriction(problem=self.problem,
                                                                   nc=self.numerical_circuit,
                                                                   theta=theta,
                                                                   F=F,
                                                                   T=T,
                                                                   ratings=branch_ratings,
                                                                   ratings_slack_from=branch_rating_slack1,
                                                                   ratings_slack_to=branch_rating_slack2,
                                                                   monitored=self.numerical_circuit.branch_data.monitor_loading,
                                                                   active=br_active)

        elif self.zonal_grouping == ZonalGrouping.All:
            flow_f = np.zeros((self.numerical_circuit.nbr, nt))
            tau = np.ones((self.numerical_circuit.nbr, nt))
            Pinj_tau = np.zeros((self.numerical_circuit.nbr, nt))

        else:
            raise ValueError()

        # set the nodal restrictions -----------------------------------------------------------------------------------
        self.nodal_restrictions = formulate_dc_nodal_power_balance(numerical_circuit=self.numerical_circuit,
                                                                   problem=self.problem,
                                                                   theta=theta,
                                                                   P=P + Pinj_tau,   # add the phase shift injections
                                                                   start_=self.start_idx,
                                                                   end_=self.end_idx)

        # if there are batteries, add the batteries --------------------------------------------------------------------
        if nb > 0:
            add_battery_discharge_restriction(problem=self.problem,
                                              SoC0=SoC0,
                                              Capacity=Capacity,
                                              Efficiency=Efficiency,
                                              Pb=Pb, E=E, dt=dt)

        if self.consider_contingencies:
            con_flow_lst, con_overload1_lst, con_overload2_lst, \
            con_br_idx = formulate_contingency(problem=self.problem,
                                               numerical_circuit=self.numerical_circuit,
                                               flow_f=flow_f,
                                               ratings=branch_ratings,
                                               LODF=self.LODF,
                                               monitor=self.numerical_circuit.branch_data.monitor_loading,
                                               lodf_tolerance=self.lodf_tolerance)
        else:
            con_flow_lst = list()
            con_br_idx = list()
            con_overload1_lst = list()
            con_overload2_lst = list()

        # maximize the power from->to ----------------------------------------------------------------------------------
        if self.maximize_inter_area_flow:
            flow_from_a1_to_a2 = formulate_inter_area_flow(numerical_circuit=self.numerical_circuit,
                                                           buses_areas_1=self.buses_areas_1,
                                                           buses_areas_2=self.buses_areas_2,
                                                           flow_f=flow_f,
                                                           hvdc_flow_f=hvdc_flow_f)

            sum_gen_area_1, sum_gen_area_2 = formulate_area_generation_summations(
                numerical_circuit=self.numerical_circuit,
                buses_areas_1=self.buses_areas_1,
                buses_areas_2=self.buses_areas_2,
                Pg=Pg)

        else:
            flow_from_a1_to_a2 = None
            sum_gen_area_1 = None
            sum_gen_area_2 = None

        # add the objective function -----------------------------------------------------------------------------------
        self.problem += get_objective_function(Pg=Pg,
                                               Pb=Pb,
                                               LSlack=load_slack,
                                               FSlack1=branch_rating_slack1,
                                               FSlack2=branch_rating_slack2,
                                               FCSlack1=con_overload1_lst,
                                               FCSlack2=con_overload2_lst,
                                               flow_from_a1_to_a2=flow_from_a1_to_a2,
                                               sum_gen_area_1=sum_gen_area_1,
                                               sum_gen_area_2=sum_gen_area_2,
                                               cost_g=cost_g,
                                               cost_b=cost_b,
                                               cost_l=cost_l,
                                               cost_br=cost_br)

        # Assign variables to keep
        # transpose them to be in the format of GridCal: time, device
        self.theta = theta.transpose()
        self.Pg = Pg.transpose()
        self.Pb = Pb.transpose()
        self.Pl = Pl.transpose()

        self.Pinj = P.transpose()

        self.phase_shift = tau.transpose()

        self.hvdc_flow = hvdc_flow_f.transpose()

        self.E = E.transpose()
        self.load_shedding = load_slack.transpose()
        self.s_from = flow_f.transpose()
        self.s_to = -flow_f.transpose()
        self.overloads = (branch_rating_slack1 + branch_rating_slack2).transpose()
        self.rating = branch_ratings.T

        self.contingency_flows_list = con_flow_lst
        self.contingency_indices_list = con_br_idx  # [(t, m, c), ...]
        self.contingency_flows_slacks_list = con_overload1_lst


if __name__ == '__main__':

    from GridCal.Engine import *

    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Lynn 5 Bus pv.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/grid_2_islands.xlsx'

    main_circuit = FileOpen(fname).open()

    # main_circuit.buses[3].controlled_generators[0].enabled_dispatch = False

    # get the power flow options from the GUI
    solver = SolverType.DC_OPF
    mip_solver = MIPSolvers.CBC
    grouping = TimeGrouping.Daily
    pf_options = PowerFlowOptions()

    options = OptimalPowerFlowOptions(solver=solver,
                                      time_grouping=grouping,
                                      mip_solver=mip_solver,
                                      power_flow_options=pf_options)

    start = 0
    end = len(main_circuit.time_profile)

    # create the OPF time series instance
    # if non_sequential:
    optimal_power_flow_time_series = OptimalPowerFlowTimeSeries(grid=main_circuit,
                                                                options=options,
                                                                start_=start,
                                                                end_=end)

    optimal_power_flow_time_series.run()

    v = optimal_power_flow_time_series.results.voltage
    print('Angles\n', np.angle(v))

    l = optimal_power_flow_time_series.results.loading
    print('Branch loading\n', l)

    g = optimal_power_flow_time_series.results.generator_power
    print('Gen power\n', g)

    pr = optimal_power_flow_time_series.results.shadow_prices
    print('Nodal prices \n', pr)

    import pandas as pd
    pd.DataFrame(optimal_power_flow_time_series.results.loading).to_excel('opf_loading.xlsx')



import numpy as np
import pandas as pd
from GridCal.Engine.Simulations.driver_template import DriverTemplate
from GridCal.Engine.Simulations.driver_types import SimulationTypes
from GridCal.Engine.Simulations.results_template import ResultsTemplate
from GridCal.Engine.Simulations.result_types import ResultTypes
from GridCal.Engine.Simulations.results_table import ResultsTable
from GridCal.Engine.Core.multi_circuit import MultiCircuit


class InputsAnalysisResults(ResultsTemplate):

    def __init__(self, grid: MultiCircuit):
        """
        Construct the analysis
        :param grid:
        """
        ResultsTemplate.__init__(self,
                                 name='Inputs analysis',
                                 available_results=[ResultTypes.ZoneAnalysis,
                                                    ResultTypes.AreaAnalysis,
                                                    ResultTypes.CountryAnalysis
                                                    ],

                                 data_variables=[])

        self.grid = grid

        self.area_names = list(set([elm.name for elm in grid.areas]))
        self.zone_names = list(set([elm.name for elm in grid.zones]))
        self.country_names = list(set([elm.name for elm in grid.countries]))

        self.gen_data = self.get_generators_df()
        self.battery_data = self.get_batteries_df()
        self.load_data = self.get_loads_df()
        self.static_gen_data = self.get_static_generators_df()

    def get_generators_df(self):
        """

        :return:
        """
        dta = list()
        for elm in self.grid.get_generators():
            dta.append([elm.name,
                            elm.P * elm.active,
                            elm.Pf,
                            elm.Snom,
                            elm.Pmin, elm.Pmax,
                            elm.Qmin, elm.Qmax,
                            elm.Vset,
                            elm.bus.zone.name,
                            elm.bus.area.name,
                            elm.bus.substation.name,
                            elm.bus.country.name])
        cols = ['Name', 'P', 'Pf',
                'Snom', 'Pmin', 'Pmax',
                'Qmin', 'Qmax', 'Vset',
                'Zone', 'Area', 'Substation', 'Country']
        return pd.DataFrame(data=dta, columns=cols)

    def get_batteries_df(self):
        """

        :return:
        """
        dta = list()
        for elm in self.grid.get_batteries():
            dta.append([elm.name,
                            elm.P * elm.active,
                            elm.Pf,
                            elm.Snom,
                            elm.Pmin, elm.Pmax,
                            elm.Qmin, elm.Qmax,
                            elm.Vset,
                            elm.bus.zone.name,
                            elm.bus.area.name,
                            elm.bus.substation.name,
                            elm.bus.country.name])
        cols = ['Name', 'P', 'Pf',
                'Snom', 'Pmin', 'Pmax',
                'Qmin', 'Qmax', 'Vset',
                'Zone', 'Area', 'Substation', 'Country']
        return pd.DataFrame(data=dta, columns=cols)

    def get_loads_df(self):
        """

        :return:
        """
        dta = list()
        for elm in self.grid.get_loads():
            dta.append([elm.name,
                        elm.P * elm.active,
                        elm.Q * elm.active,
                        elm.bus.zone.name,
                        elm.bus.area.name,
                        elm.bus.substation.name,
                        elm.bus.country.name])
        cols = ['Name', 'P', 'Q',
                'Zone', 'Area', 'Substation', 'Country']
        return pd.DataFrame(data=dta, columns=cols)

    def get_static_generators_df(self):
        """

        :return:
        """
        dta = list()
        for elm in self.grid.get_static_generators():
            dta.append([elm.name,
                        elm.P * elm.active,
                        elm.Q * elm.active,
                        elm.bus.zone.name,
                        elm.bus.area.name,
                        elm.bus.substation.name,
                        elm.bus.country.name])
        cols = ['Name', 'P', 'Q',
                'Zone', 'Area', 'Substation', 'Country']
        return pd.DataFrame(data=dta, columns=cols)

    def group_by(self, group: str):
        """
        Return a DataFrame grouped by Area, Zone or Country
        :param group: "Area", "Zone" or "Country"
        :return: Group DataFrame
        """
        if group == 'Area':
            labels = self.area_names
        elif group == 'Zone':
            labels = self.zone_names
        elif group == 'Country':
            labels = self.country_names
        else:
            raise Exception('Unknown grouping:' + str(group))

        n = len(labels)
        cols_gen = ['P', 'Pmin', 'Pmax', 'Qmin', 'Qmax']
        cols_load = ['P', 'Q']
        cols = ['P', 'Pgen', 'Pload', 'Pbatt', 'Pstagen', 'Pmin', 'Pmax', 'Q', 'Qmin', 'Qmax']
        df = pd.DataFrame(data=np.zeros((n, len(cols))), columns=cols, index=labels)

        if len(self.gen_data):
            df2 = self.gen_data.groupby(group).sum()
            df[cols_gen] += df2[cols_gen]
            df['Pgen'] = df2['P']

        if len(self.battery_data):
            df2 = self.battery_data.groupby(group).sum()
            df[cols_gen] += df2[cols_gen]
            df['Pbatt'] = df2['P']

        if len(self.load_data):
            df2 = self.load_data.groupby(group).sum()
            df[cols_load] -= df2[cols_load]
            df['Pload'] = df2['P']

        if len(self.static_gen_data):
            df2 = self.static_gen_data.groupby(group).sum()
            df[cols_load] += df2[cols_load]
            df['Pstagen'] = df2['P']

        df.fillna(0, inplace=True)

        return df

    def mdl(self, result_type) -> "ResultsTable":
        """
        Plot the results
        :param result_type: type of results (string)
        :return: DataFrame of the results
                (or None if the result was not understood)
        """

        columns = [result_type.value[0]]

        if result_type == ResultTypes.ZoneAnalysis:
            df = self.group_by('Zone')
            columns = df.columns.values
            labels = df.index.values
            y = df.values
            y_label = ''
            title = result_type.value[0]

        elif result_type == ResultTypes.AreaAnalysis:
            df = self.group_by('Area')
            columns = df.columns.values
            labels = df.index.values
            y = df.values
            y_label = ''
            title = result_type.value[0]

        elif result_type == ResultTypes.CountryAnalysis:
            df = self.group_by('Country')
            columns = df.columns.values
            labels = df.index.values
            y = df.values
            y_label = ''
            title = result_type.value[0]

        else:
            labels = []
            y = np.zeros(0)
            y_label = '(MW)'
            title = ''

        mdl = ResultsTable(data=y,
                           index=labels,
                           columns=columns,
                           title=title,
                           ylabel=y_label,
                           xlabel='',
                           units=y_label)
        return mdl


class InputsAnalysisDriver(DriverTemplate):
    name = 'Inputs Analysis'
    tpe = SimulationTypes.InputsAnalysis_run

    def __init__(self, grid: MultiCircuit):
        """
        InputsAnalysisDriver class constructor
        :param grid: MultiCircuit instance
        """
        DriverTemplate.__init__(self, grid=grid)

        self.results = InputsAnalysisResults(grid=grid)

    def get_steps(self):
        """

        :return:
        """
        return list()

    def run(self):
        """
        Pack run_pf for the QThread
        :return:
        """
        pass

    def cancel(self):
        self.__cancel__ = True


if __name__ == '__main__':
    from GridCal.Engine.IO.file_handler import FileOpen

    # fname = '/home/santi/Documentos/Git/GitHub/GridCal/Grids_and_profiles/grids/IEEE14 - ntc areas.gridcal'
    # fname = '/home/santi/Documentos/Git/GitHub/GridCal/Grids_and_profiles/grids/IEEE 118 Bus - ntc_areas.gridcal'
    fname = r'C:\Users\SPV86\Documents\Git\GitHub\GridCal\Grids_and_profiles\grids\IEEE 118 Bus - ntc_areas.gridcal'

    main_circuit = FileOpen(fname).open()

    drv = InputsAnalysisDriver(grid=main_circuit)

    mdl = drv.results.mdl(ResultTypes.AreaAnalysis)

    df = mdl.to_df()

    print(df)




# -*- coding: utf-8 -*-
# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import numpy as np
from enum import Enum
from scipy.sparse import csr_matrix, csc_matrix


class SparseSolver(Enum):
    BLAS_LAPACK = 'Blas/Lapack'
    ILU = 'ILU'
    KLU = 'KLU'
    SuperLU = 'SuperLU'
    Pardiso = 'Pardiso'
    GMRES = 'GMRES'
    UMFPACK = 'UmfPack'

    def __str__(self):
        return self.value


# list of available linear algebra frameworks
available_sparse_solvers = list()


try:
    import cvxopt
    from cvxoptklu import klu
    spsolve = klu.linsolve

    available_sparse_solvers.append(SparseSolver.KLU)
except ImportError:
    pass
    # print(SparseSolver.KLU.value + ' failed')


try:
    from scipy.sparse.linalg import spsolve as scipy_spsolve, splu, spilu, gmres
    available_sparse_solvers.append(SparseSolver.BLAS_LAPACK)
    available_sparse_solvers.append(SparseSolver.ILU)
    available_sparse_solvers.append(SparseSolver.SuperLU)
    available_sparse_solvers.append(SparseSolver.GMRES)
except ImportError:
    pass
    # print(SparseSolver.BLAS_LAPACK.value + ' failed')


try:
    from pypardiso import spsolve as pardiso_spsolve

    available_sparse_solvers.append(SparseSolver.Pardiso)
except ImportError:
    pass
    # print(SparseSolver.Pardiso.value + ' failed')

try:
    from scikits.umfpack import spsolve_umfpack

    available_sparse_solvers.append(SparseSolver.UMFPACK)
except ImportError:
    pass
    # print(SparseSolver.UMFPACK.value + ' failed')


preferred_type = SparseSolver.KLU

if preferred_type not in available_sparse_solvers:
    if len(available_sparse_solvers) > 0:
        preferred_type = available_sparse_solvers[0]
        # print('Falling back to', preferred_type)
    else:
        raise Exception('No linear algebra solver!!!! GridCal cannot work without one.')
# print('Using', preferred_type)


def get_sparse_type(solver_type: SparseSolver = preferred_type):
    """
    GEt sparse matrix type matching the selected sparse linear systems solver
    :param solver_type:
    :return: sparse matrix type
    """
    if solver_type in [SparseSolver.BLAS_LAPACK, SparseSolver.Pardiso, SparseSolver.GMRES]:
        return csr_matrix

    elif solver_type in [SparseSolver.KLU, SparseSolver.SuperLU, SparseSolver.ILU, SparseSolver.UMFPACK]:
        return csc_matrix

    else:
        raise Exception('Unknown solver' + str(solver_type))


def super_lu_linsolver(A, b):
    """
    SuperLU wrapper function for linear system solve A x = b
    :param A: System matrix
    :param b: right hand side
    :return: solution
    """
    return splu(A).solve(b)


def ilu_linsolver(A, b):
    """
    ILU wrapper function for linear system solve A x = b
    :param A: System matrix
    :param b: right hand side
    :return: solution
    """
    return spilu(A).solve(b)


def klu_linsolve(A, b):
    """
    KLU wrapper function for linear system solve A x = b
    :param A: System matrix
    :param b: right hand side
    :return: solution
    """
    A2 = A.tocoo()
    A_cvxopt = cvxopt.spmatrix(A2.data, A2.row, A2.col, A2.shape, 'd')
    x = cvxopt.matrix(b)
    klu.linsolve(A_cvxopt, x)
    return np.array(x)[:, 0]


def gmres_linsolve(A, b):
    """

    :param A:
    :param b:
    :return:
    """
    x, info = gmres(A, b)
    return x


def umfpack_linsolve(A, b):
    """

    :param A:
    :param b:
    :return:
    """
    return spsolve_umfpack(A, b)


def get_linear_solver(solver_type: SparseSolver = preferred_type):
    """
    Privide the chosen linear solver_type function pointer to solver_type linear systems of the type A x = b, with x = f(A,b)
    :param solver_type: SparseSolver option
    :return: function pointer f(A, b)
    """
    if solver_type == SparseSolver.BLAS_LAPACK:
        return scipy_spsolve

    elif solver_type == SparseSolver.KLU:
        return klu_linsolve

    elif solver_type == SparseSolver.SuperLU:
        return super_lu_linsolver

    elif solver_type == SparseSolver.Pardiso:
        return pardiso_spsolve

    elif solver_type == SparseSolver.ILU:
        return ilu_linsolver

    elif solver_type == SparseSolver.GMRES:
            return gmres_linsolve

    elif solver_type == SparseSolver.UMFPACK:
            return umfpack_linsolve


if __name__ == '__main__':

    import time
    import scipy.sparse as sp

    solver_types = [SparseSolver.BLAS_LAPACK,
                    SparseSolver.KLU,
                    SparseSolver.SuperLU,
                    SparseSolver.ILU,
                    SparseSolver.UMFPACK
                    ]

    for solver_type_ in solver_types:
        start = time.time()
        repetitions = 50
        n = 4000
        np.random.seed(0)

        sparse = get_sparse_type(solver_type_)
        solver = get_linear_solver(solver_type_)

        for r in range(repetitions):
            A = sparse(sp.rand(n, n, 0.01)) + sp.diags(np.random.rand(n) * 10.0, shape=(n, n))
            b = np.random.rand(n)
            x = solver(A, b)

        end = time.time()
        dt = end - start
        print(solver_type_, '  total', dt, 's, average:', dt / repetitions, 's')



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import numpy as np
import scipy.sparse as sp
import GridCal.Engine.Core.topology as tp


class LinesData:

    def __init__(self, nline, nbus, ntime=1):
        """

        :param nline:
        :param nbus:
        """
        self.nline = nline
        self.ntime = ntime

        self.line_names = np.zeros(nline, dtype=object)
        self.line_active = np.zeros((nline, ntime), dtype=int)
        self.line_R = np.zeros(nline, dtype=float)
        self.line_X = np.zeros(nline, dtype=float)
        self.line_B = np.zeros(nline, dtype=float)

        self.C_line_bus = sp.lil_matrix((nline, nbus), dtype=int)  # this ons is just for splitting islands

    def slice(self, elm_idx, bus_idx, time_idx=None):
        """

        :param elm_idx:
        :param bus_idx:
        :return:
        """

        if time_idx is None:
            idx = elm_idx
        else:
            idx = np.ix_(elm_idx, time_idx)

        data = LinesData(nline=len(elm_idx), nbus=len(bus_idx))

        data.line_active = self.line_active[idx]

        data.line_names = self.line_names[elm_idx]

        data.line_R = self.line_R[elm_idx]
        data.line_X = self.line_X[elm_idx]
        data.line_B = self.line_B[elm_idx]
        data.C_line_bus = self.C_line_bus[np.ix_(elm_idx, bus_idx)]

        return data

    def get_island(self, bus_idx, t_idx=0):
        """
        Get the elements of the island given the bus indices
        :param bus_idx: list of bus indices
        :return: list of line indices of the island
        """
        if self.nline:
            return tp.get_elements_of_the_island(self.C_line_bus, bus_idx,
                                                 active=self.line_active[:, t_idx])
        else:
            return np.zeros(0, dtype=int)

    def __len__(self):
        return self.nline




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.PowerFlow.power_flow_driver import PowerFlowResults
from GridCal.Engine.Simulations.Dynamics.dynamic_modules import dynamic_simulation
from GridCal.Engine.Simulations.driver_template import DriverTemplate

########################################################################################################################
# Transient stability
########################################################################################################################


class TransientStabilityOptions:

    def __init__(self, h=0.001, t_sim=15, max_err=0.0001, max_iter=25):

        # step length (s)
        self.h = h

        # simulation time (s)
        self.t_sim = t_sim

        # Maximum error in network iteration (voltage mismatches)
        self.max_err = max_err

        # Maximum number of network iterations
        self.max_iter = max_iter


class TransientStability(DriverTemplate):

    def __init__(self, grid: MultiCircuit, options: TransientStabilityOptions, pf_res: PowerFlowResults):
        """
        TimeSeries constructor
        @param grid: MultiCircuit instance
        @param options: PowerFlowOptions instance
        """
        DriverTemplate.__init__(self, grid=grid)

        self.grid = grid

        self.options = options

        self.pf_res = pf_res

        self.results = None

    def get_steps(self):
        """
        Get time steps list of strings
        """
        return list()

    def status(self, txt, progress):
        """
        Emit status
        :param txt: text to display
        :param progress: progress 0-100
        """
        self.progress_signal.emit(progress)
        self.progress_text.emit(txt)

    def run(self):
        """
        Run transient stability
        """
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Running transient stability...')

        # print('Compiling...', end='')
        numerical_circuit = self.grid.compile_snapshot()
        calculation_inputs = numerical_circuit.compute()

        for calculation_input in calculation_inputs:

            dynamic_devices = calculation_input.get_generators()
            bus_indices = [calculation_input.buses_dict[elm.bus] for elm in dynamic_devices]

            res = dynamic_simulation(n=len(calculation_input.buses),
                                     Vbus=self.pf_res.voltage[calculation_input.original_bus_idx],
                                     Sbus=self.pf_res.Sbus[calculation_input.original_bus_idx],
                                     Ybus=calculation_input.Ybus,
                                     Sbase=calculation_input.Sbase,
                                     fBase=calculation_input.fBase,
                                     t_sim=self.options.t_sim,
                                     h=self.options.h,
                                     dynamic_devices=dynamic_devices,
                                     bus_indices=bus_indices,
                                     callback=self.status)

        self.results = res



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from GridCal.Engine.Simulations.Topology.topology_driver import *



# PuLP : Python LP Modeler
# Version 1.4.2

# Copyright (c) 2002-2005, Jean-Sebastien Roy (js@jeannot.org)
# Modifications Copyright (c) 2007- Stuart Anthony Mitchell (s.mitchell@auckland.ac.nz)
# $Id:solvers.py 1791 2008-04-23 22:54:34Z smit023 $

# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:

# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE."""

"""
This file contains the solver classes for PuLP
Note that the solvers that require a compiled extension may not work in
the current version
"""

import os
import sys
try:
    from time import clock
except ImportError:
    from time import perf_counter as clock
from uuid import uuid4
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
try:
    Parser = configparser.ConfigParser
except AttributeError:
    Parser = configparser.SafeConfigParser
from GridCal.ThirdParty.pulp import sparse
import warnings
from tempfile import mktemp
from GridCal.ThirdParty.pulp.constants import *

import logging
log = logging.getLogger(__name__)

if os.name == "posix" and sys.version_info[0] < 3:
    try:
        import subprocess32 as subprocess
    except ImportError:
        log.debug("Thread-safe subprocess32 module not found! "
                  "Using unsafe built-in subprocess module instead.")
        import subprocess
else:
    import subprocess


class PulpSolverError(PulpError):
    """
    Pulp Solver-related exceptions
    """
    pass


# import configuration information
def initialize(filename, operating_system='linux', arch='64'):
    """
    reads the configuration file to initialise the module
    :param filename:
    :param operating_system:
    :param arch:
    :return:
    """

    here = os.path.dirname(filename)

    config = Parser({'here': here, 'os': operating_system, 'arch': arch})
    config.read(filename)

    try:
        cplex_dll_path = config.get("locations", "CplexPath")
    except configparser.Error:
        cplex_dll_path = 'libcplex110.so'
    try:
        try:
            ilm_cplex_license = config.get("licenses", "ilm_cplex_license").decode("string-escape").replace('"','')
        except AttributeError:
            ilm_cplex_license = config.get("licenses", "ilm_cplex_license").replace('"', '')

    except configparser.Error:
        ilm_cplex_license = ''

    try:
        ilm_cplex_license_signature = config.getint("licenses", "ilm_cplex_license_signature")
    except configparser.Error:
        ilm_cplex_license_signature = 0

    try:
        coinMP_path = config.get("locations", "CoinMPPath").split(', ')
    except configparser.Error:
        coinMP_path = ['libCoinMP.so']

    try:
        gurobi_path = config.get("locations", "GurobiPath")
    except configparser.Error:
        gurobi_path = '/opt/gurobi201/linux32/lib/python2.5'

    try:
        cbc_path = config.get("locations", "CbcPath")
    except configparser.Error:
        cbc_path = 'cbc'

    try:
        glpk_path = config.get("locations", "GlpkPath")
    except configparser.Error:
        glpk_path = 'glpsol'

    try:
        pulp_cbc_path = config.get("locations", "PulpCbcPath")
    except configparser.Error:
        pulp_cbc_path = 'cbc'

    try:
        pulp_highs_path = config.get("locations", "PulpHiGHSPath")
    except configparser.Error:
        pulp_highs_path = 'highs'

    try:
        scip_path = config.get("locations", "ScipPath")
    except configparser.Error:
        scip_path = 'scip'

    for i, path in enumerate(coinMP_path):
        if not os.path.dirname(path):
            # if no pathname is supplied assume the file is in the same directory
            coinMP_path[i] = os.path.join(os.path.dirname(config_filename), path)

    return cplex_dll_path, ilm_cplex_license, ilm_cplex_license_signature,\
        coinMP_path, gurobi_path, cbc_path, glpk_path, pulp_cbc_path, scip_path, pulp_highs_path


# pick up the correct config file depending on operating system
PULPCFGFILE = "pulp.cfg"
is_64bits = sys.maxsize > 2**32
if is_64bits:
    arch = '64'
else:
    arch = '32'
operating_system = None
if sys.platform in ['win32', 'cli']:
    operating_system = 'win'
    PULPCFGFILE += ".win"
elif sys.platform in ['darwin']:
    operating_system = "osx"
    arch = '64'
    PULPCFGFILE += ".osx"
else:
    operating_system = "linux"
    PULPCFGFILE += ".linux"

if __name__ != '__main__':
    DIRNAME = os.path.dirname(__file__)
    config_filename = os.path.join(DIRNAME, PULPCFGFILE)
else:
    # run as a script
    from GridCal.ThirdParty.pulp import __file__ as fname
    DIRNAME = os.path.dirname(fname)
    config_filename = os.path.join(DIRNAME, PULPCFGFILE)


cplex_dll_path, \
 ilm_cplex_license, \
 ilm_cplex_license_signature, \
 coinMP_path,\
 gurobi_path, \
 cbc_path, \
 glpk_path, \
 pulp_cbc_path, \
 scip_path, pulp_highs_path = initialize(config_filename, operating_system, arch)


try:

    import ctypes

    def ctypesArrayFill(myList, type=ctypes.c_double):
        """
        Creates a c array with ctypes from a python list
        type is the type of the c array
        """
        ctype= type * len(myList)
        cList = ctype()
        for i, elem in enumerate(myList):
            cList[i] = elem
        return cList

except(ImportError):

    def ctypesArrayFill(myList, type = None):
        return None


# See later for LpSolverDefault definition
class LpSolver:
    """A generic LP Solver"""

    def __init__(self, mip=True, msg=True, options=None, *args, **kwargs):
        self.mip = mip
        self.msg = msg

        if options is None:
            self.options = list()
        else:
            self.options = options

    def available(self):
        """True if the solver is available"""
        raise NotImplementedError

    def actualSolve(self, lp):
        """Solve a well formulated lp problem"""
        raise NotImplementedError

    def actualResolve(self, lp, **kwargs):
        """
        uses existing problem information and solves the problem
        If it is not implelemented in the solver
        just solve again
        """
        self.actualSolve(lp, **kwargs)

    def copy(self):
        """Make a copy of self"""

        aCopy = self.__class__()
        aCopy.mip = self.mip
        aCopy.msg = self.msg
        aCopy.options = self.options
        return aCopy

    def solve(self, lp):
        """Solve the problem lp"""
        # Always go through the solve method of LpProblem
        return lp.solve(self)

    # TODO: Not sure if this code should be here or in a child class
    def getCplexStyleArrays(self, lp,
                            senseDict={LpConstraintEQ: "E", LpConstraintLE: "L", LpConstraintGE: "G"},
                            LpVarCategories={LpContinuous: "C", LpInteger: "I"},
                            LpObjSenses={LpMaximize: -1, LpMinimize: 1},
                            infBound=1e20):
        """
        returns the arrays suitable to pass to a cdll Cplex
        or other solvers that are similar
        Copyright (c) Stuart Mitchell 2007
        :param lp:
        :param senseDict:
        :param LpVarCategories:
        :param LpObjSenses:
        :param infBound:
        :return:
        """

        rangeCount = 0
        variables = list(lp.variables())
        numVars = len(variables)

        # associate each variable with a ordinal
        self.v2n = dict(((variables[i], i) for i in range(numVars)))
        self.vname2n = dict(((variables[i].name, i) for i in range(numVars)))
        self.n2v = dict((i, variables[i]) for i in range(numVars))

        # objective values
        objSense = LpObjSenses[lp.sense]

        NumVarDoubleArray = ctypes.c_double * numVars

        objectCoeffs = NumVarDoubleArray()

        for v, val in lp.objective.items():
            objectCoeffs[self.v2n[v]] = val

        # values for variables
        objectConst = ctypes.c_double(0.0)
        NumVarStrArray = ctypes.c_char_p * numVars
        colNames = NumVarStrArray()
        lowerBounds = NumVarDoubleArray()
        upperBounds = NumVarDoubleArray()
        initValues = NumVarDoubleArray()
        for v in lp.variables():
            colNames[self.v2n[v]] = str(v.name)
            initValues[self.v2n[v]] = 0.0
            if v.lowBound != None:
                lowerBounds[self.v2n[v]] = v.lowBound
            else:
                lowerBounds[self.v2n[v]] = -infBound
            if v.upBound != None:
                upperBounds[self.v2n[v]] = v.upBound
            else:
                upperBounds[self.v2n[v]] = infBound
        # values for constraints
        numRows = len(lp.constraints)
        NumRowDoubleArray = ctypes.c_double * numRows
        NumRowStrArray = ctypes.c_char_p * numRows
        NumRowCharArray = ctypes.c_char * numRows
        rhsValues = NumRowDoubleArray()
        rangeValues = NumRowDoubleArray()
        rowNames = NumRowStrArray()
        rowType = NumRowCharArray()

        self.c2n = {}

        self.n2c = {}

        i = 0
        for c in lp.constraints:
            rhsValues[i] = -lp.constraints[c].constant
            # for ranged constraints a<= constraint >=b
            rangeValues[i] = 0.0
            rowNames[i] = str(c)
            rowType[i] = senseDict[lp.constraints[c].sense]
            self.c2n[c] = i
            self.n2c[i] = c
            i = i + 1
        # return the coefficient matrix as a series of vectors
        coeffs = lp.coefficients()
        sparseMatrix = sparse.Matrix(list(range(numRows)), list(range(numVars)))
        for var, row, coeff in coeffs:
            sparseMatrix.add(self.c2n[row], self.vname2n[var], coeff)
        (numels, mystartsBase, mylenBase, myindBase, myelemBase) = sparseMatrix.col_based_arrays()

        elemBase = ctypesArrayFill(myelemBase, ctypes.c_double)

        indBase = ctypesArrayFill(myindBase, ctypes.c_int)

        startsBase = ctypesArrayFill(mystartsBase, ctypes.c_int)

        lenBase = ctypesArrayFill(mylenBase, ctypes.c_int)

        # MIP Variables
        NumVarCharArray = ctypes.c_char * numVars

        columnType = NumVarCharArray()
        if lp.isMIP():
            for v in lp.variables():
                columnType[self.v2n[v]] = LpVarCategories[v.cat]

        self.addedVars = numVars

        self.addedRows = numRows

        return (numVars, numRows, numels, rangeCount,
                objSense, objectCoeffs, objectConst,
                rhsValues, rangeValues, rowType, startsBase, lenBase, indBase,
                elemBase, lowerBounds, upperBounds, initValues, colNames,
                rowNames, columnType, self.n2v, self.n2c)


class LpSolver_CMD(LpSolver):
    """A generic command line LP Solver"""

    def __init__(self, path=None, keepFiles=0, mip=1, msg=1, options=None):

        if options is None:
            self.options = list()
        else:
            self.options = options

        LpSolver.__init__(self, mip, msg, options)
        if path is None:
            self.path = self.defaultPath()
        else:
            self.path = path
        self.keepFiles = keepFiles
        self.setTmpDir()

    def copy(self):
        """Make a copy of self"""

        aCopy = LpSolver.copy(self)
        aCopy.path = self.path
        aCopy.keepFiles = self.keepFiles
        aCopy.tmpDir = self.tmpDir
        return aCopy

    def setTmpDir(self):
        """Set the tmpDir attribute to a reasonable location for a temporary
        directory"""
        if os.name != 'nt':
            # On unix use /tmp by default
            self.tmpDir = os.environ.get("TMPDIR", "/tmp")
            self.tmpDir = os.environ.get("TMP", self.tmpDir)
        else:
            # On Windows use the current directory
            self.tmpDir = os.environ.get("TMPDIR", "")
            self.tmpDir = os.environ.get("TMP", self.tmpDir)
            self.tmpDir = os.environ.get("TEMP", self.tmpDir)
        if not os.path.isdir(self.tmpDir):
            self.tmpDir = ""
        elif not os.access(self.tmpDir, os.F_OK + os.W_OK):
            self.tmpDir = ""

    def create_tmp_files(self, name, *args):
        if self.keepFiles:
            prefix = name
        else:
            prefix = os.path.join(self.tmpDir, uuid4().hex)
        return ("%s-pulp.%s" % (prefix, n) for n in args)

    def delete_tmp_files(self, *args):
        if self.keepFiles:
            return
        for file in args:
            try:
                os.remove(file)
            except:
                pass

    def defaultPath(self):
        raise NotImplementedError

    def executableExtension(name):
        if os.name != 'nt':
            return name
        else:
            return name + ".exe"

    executableExtension = staticmethod(executableExtension)

    def executable(command):
        """Checks that the solver command is executable,
        And returns the actual path to it."""
        import os
        if os.path.isabs(command):
            if os.path.exists(command) and os.access(command, os.X_OK):
                return command
        for path in os.environ["PATH"].split(os.pathsep):
            new_path = os.path.join(path, command)
            if os.path.exists(new_path) and os.access(new_path, os.X_OK):
                return os.path.join(path, command)

        print('Unable to find the command', command, 'in the path', os.environ["PATH"])
        return False

    executable = staticmethod(executable)




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import sys
import os
import numpy as np
import pandas as pd
from typing import List
import networkx as nx
from warnings import warn


from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtSvg import QSvgGenerator
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Devices.bus import Bus
from GridCal.Engine.Devices.line import Line
from GridCal.Engine.Devices.dc_line import DcLine
from GridCal.Engine.Devices.transformer import Transformer2W
from GridCal.Engine.Devices.vsc import VSC
from GridCal.Engine.Devices.upfc import UPFC
from GridCal.Engine.Devices.hvdc_line import HvdcLine
from GridCal.Engine.Devices.transformer3w import Transformer3W
from GridCal.Gui.GridEditorWidget.terminal_item import TerminalItem
from GridCal.Gui.GridEditorWidget.bus_graphics import BusGraphicItem
from GridCal.Gui.GridEditorWidget.line_graphics import LineGraphicItem
from GridCal.Gui.GridEditorWidget.dc_line_graphics import DcLineGraphicItem
from GridCal.Gui.GridEditorWidget.transformer2w_graphics import TransformerGraphicItem
from GridCal.Gui.GridEditorWidget.hvdc_graphics import HvdcGraphicItem
from GridCal.Gui.GridEditorWidget.vsc_graphics import VscGraphicItem
from GridCal.Gui.GridEditorWidget.upfc_graphics import UpfcGraphicItem
from GridCal.Gui.GridEditorWidget.transformer3w_graphics import Transformer3WGraphicItem
from GridCal.Engine.Simulations.driver_types import SimulationTypes
from matplotlib import pyplot as plt

'''
Dependencies:

GridEditor
 |
  - EditorGraphicsView (Handles the drag and drop)
 |   |
  ---- DiagramScene
        |
         - MultiCircuit (Calculation engine)
        |
         - Graphic Objects: (BusGraphicItem, BranchGraphicItem, LoadGraphicItem, ...)


The graphic objects need to call the API objects and functions inside the MultiCircuit instance.
To do this the graphic objects call "parent.circuit.<function or object>"
'''

def toQBytesArray(val: str):
    data = QByteArray()
    stream = QDataStream(data, QIODevice.WriteOnly)
    stream.writeQString(val)
    return data


class EditorGraphicsView(QGraphicsView):

    def __init__(self, scene, parent=None, editor=None):
        """
        Editor where the diagram is displayed
        @param scene: DiagramScene object
        @param parent:
        @param editor:
        """
        QGraphicsView.__init__(self, scene, parent)
        self._zoom = 0
        self.setDragMode(QGraphicsView.RubberBandDrag)
        self.setRubberBandSelectionMode(Qt.IntersectsItemShape)
        self.setMouseTracking(True)
        self.setInteractive(True)
        self.scene_ = scene
        self.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform)
        self.editor = editor
        self.setAlignment(Qt.AlignCenter)

    def adapt_map_size(self):
        w = self.size().width()
        h = self.size().height()
        print('EditorGraphicsView size: ', w, h)
        self.map.change_size(w, h)

    def dragEnterEvent(self, event):
        """

        @param event:
        @return:
        """
        if event.mimeData().hasFormat('component/name'):
            event.accept()

    def dragMoveEvent(self, event):
        """
        Move element
        @param event:
        @return:
        """
        if event.mimeData().hasFormat('component/name'):
            event.accept()

    def dropEvent(self, event):
        """
        Create an element
        @param event:
        @return:
        """
        if event.mimeData().hasFormat('component/name'):
            obj_type = event.mimeData().data('component/name')
            elm = None
            bus_data = toQBytesArray('Bus')
            tr3w_data = toQBytesArray('3W-Transformer')

            if bus_data == obj_type:
                name = 'Bus ' + str(len(self.scene_.circuit.buses))

                obj = Bus(name=name,
                          area=self.scene_.circuit.areas[0],
                          zone=self.scene_.circuit.zones[0],
                          substation=self.scene_.circuit.substations[0],
                          country=self.scene_.circuit.countries[0])

                elm = BusGraphicItem(diagramScene=self.scene(), name=name, editor=self.editor, bus=obj)
                obj.graphic_obj = elm
                self.scene_.circuit.add_bus(obj)  # weird but it's the only way to have graphical-API communication

            elif tr3w_data == obj_type:
                name = "Transformer 3-windings" + str(len(self.scene_.circuit.transformers3w))
                obj = Transformer3W(name=name)
                elm = Transformer3WGraphicItem(diagramScene=self.scene(), editor=self.editor, elm=obj)
                obj.graphic_obj = elm
                self.scene_.circuit.add_transformer3w(obj)  # weird but it's the only way to have graphical-API communication
                print('3w transformer dropped')

            if elm is not None:
                elm.setPos(self.mapToScene(event.pos()))
                self.scene_.addItem(elm)

    def wheelEvent(self, event):
        """
        Zoom
        @param event:
        @return:
        """
        self.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)

        # Scale the view / do the zoom
        scale_factor = 1.15
        # print(event.angleDelta().x(), event.angleDelta().y(), event.angleDelta().manhattanLength() )
        if event.angleDelta().y() > 0:
            # Zoom in
            self.scale(scale_factor, scale_factor)

        else:
            # Zooming out
            self.scale(1.0 / scale_factor, 1.0 / scale_factor)

    def add_bus(self, bus: Bus, explode_factor=1.0):
        """
        Add bus
        Args:
            bus: GridCal Bus object
            explode_factor: factor to position the node
        """
        elm = BusGraphicItem(diagramScene=self.scene(), name=bus.name, editor=self.editor, bus=bus)
        x = int(bus.x * explode_factor)
        y = int(bus.y * explode_factor)
        elm.setPos(self.mapToScene(QPoint(x, y)))
        self.scene_.addItem(elm)
        return elm


class LibraryModel(QStandardItemModel):
    """
    Items model to host the draggable icons
    This is the list of draggable items
    """

    def __init__(self, parent=None):
        """
        Items model to host the draggable icons
        @param parent:
        """
        QStandardItemModel.__init__(self, parent)

    def mimeTypes(self):
        """

        @return:
        """
        return ['component/name']

    def mimeData(self, idxs):
        """

        @param idxs:
        @return:
        """
        mimedata = QMimeData()
        for idx in idxs:
            if idx.isValid():
                txt = self.data(idx, Qt.DisplayRole)

                data = QByteArray()
                stream = QDataStream(data, QIODevice.WriteOnly)
                stream.writeQString(txt)

                mimedata.setData('component/name', data)
        return mimedata


class DiagramScene(QGraphicsScene):

    def __init__(self, parent=None, circuit: MultiCircuit = None):
        """

        @param parent:
        """
        super(DiagramScene, self).__init__(parent)
        self.parent_ = parent
        self.circuit = circuit
        self.results_dictionary = dict()

    def set_results_to_plot(self, all_threads):
        self.results_dictionary = {thr.tpe: thr for thr in all_threads if thr is not None}

    def plot_bus(self, i, api_object):
        """
        Plot branch results
        :param i: branch index (not counting HVDC lines because those are not real branches)
        :param api_object: API object
        :return:
        """
        fig = plt.figure(figsize=(12, 8))
        ax_1 = fig.add_subplot(211)
        # ax_2 = fig.add_subplot(212)

        # set time
        x = self.circuit.time_profile

        if x is not None:
            if len(x) > 0:

                # search available results
                power_data = api_object.get_active_injection_profiles_dictionary()
                voltage = dict()

                for key, driver in self.results_dictionary.items():
                    if hasattr(driver, 'results'):
                        if driver.results is not None:
                            if key == SimulationTypes.TimeSeries_run:
                                voltage[key] = np.abs(driver.results.voltage[:, i])

                # injections
                if len(power_data.keys()):
                    df = pd.DataFrame(data=power_data, index=x)
                    ax_1.set_title('Power', fontsize=14)
                    ax_1.set_ylabel('Injections [MW]', fontsize=11)
                    df.plot.area(ax=ax_1)

                # voltage
                if len(voltage.keys()):
                    ax_2 = fig.add_subplot(212, sharex=ax_1)
                    df = pd.DataFrame(data=voltage, index=x)
                    ax_2.set_title('Time', fontsize=14)
                    ax_2.set_ylabel('Voltage [p.u]', fontsize=11)
                    df.plot(ax=ax_2)

                plt.legend()
                fig.suptitle(api_object.name, fontsize=20)

                # plot the profiles
                plt.show()

    def plot_branch(self, i, api_object):
        """
        Plot branch results
        :param i: branch index (not counting HVDC lines because those are not real branches)
        :param api_object: API object
        :return:
        """
        fig = plt.figure(figsize=(12, 8))
        ax_1 = fig.add_subplot(211)
        ax_2 = fig.add_subplot(212)

        # set time
        x = self.circuit.time_profile
        x_cl = x

        if x is not None:
            if len(x) > 0:

                p = np.arange(len(x)).astype(float) / len(x)

                # search available results
                power_data = dict()
                loading_data = dict()
                loading_st_data = None
                loading_clustering_data = None
                power_clustering_data = None

                for key, driver in self.results_dictionary.items():
                    if hasattr(driver, 'results'):
                        if driver.results is not None:
                            if key == SimulationTypes.TimeSeries_run:
                                power_data[key.value] = driver.results.Sf.real[:, i]
                                loading_data[key.value] = np.sort(np.abs(driver.results.loading.real[:, i] * 100.0))

                            elif key == SimulationTypes.ClusteringTimeSeries_run:
                                x_cl = x[driver.sampled_time_idx]
                                power_clustering_data = driver.results.Sf.real[:, i]
                                loading_clustering_data = np.sort(np.abs(driver.results.loading.real[:, i] * 100.0))

                            elif key == SimulationTypes.LinearAnalysis_TS_run:
                                power_data[key.value] = driver.results.Sf.real[:, i]
                                loading_data[key.value] = np.sort(np.abs(driver.results.loading.real[:, i] * 100.0))

                            # elif key == SimulationTypes.NetTransferCapacityTS_run:
                            #     power_data[key.value] = driver.results.atc[:, i]
                            #     atc_perc = driver.results.atc[:, i] / (api_object.rate_prof + 1e-9)
                            #     loading_data[key.value] = np.sort(np.abs(atc_perc * 100.0))

                            elif key == SimulationTypes.ContingencyAnalysisTS_run:
                                power_data[key.value] = driver.results.worst_flows.real[:, i]
                                loading_data[key.value] = np.sort(np.abs(driver.results.worst_loading.real[:, i] * 100.0))

                            elif key == SimulationTypes.OPFTimeSeries_run:
                                power_data[key.value] = driver.results.Sf.real[:, i]
                                loading_data[key.value] = np.sort(np.abs(driver.results.loading.real[:, i] * 100.0))

                            elif key == SimulationTypes.StochasticPowerFlow:
                                loading_st_data = np.sort(np.abs(driver.results.loading_points.real[:, i] * 100.0))

                # add the rating
                # power_data['Rates+'] = api_object.rate_prof
                # power_data['Rates-'] = -api_object.rate_prof

                # loading
                if len(loading_data.keys()):
                    df = pd.DataFrame(data=loading_data, index=p)
                    ax_1.set_title('Probability x < value', fontsize=14)
                    ax_1.set_ylabel('Loading [%]', fontsize=11)
                    df.plot(ax=ax_1)

                if loading_clustering_data is not None:
                    p_st = np.arange(len(loading_clustering_data)).astype(float) / len(loading_clustering_data)
                    df = pd.DataFrame(data=loading_clustering_data,
                                      index=p_st,
                                      columns=[SimulationTypes.ClusteringTimeSeries_run.value])
                    ax_1.set_title('Probability x < value', fontsize=14)
                    ax_1.set_ylabel('Loading [%]', fontsize=11)
                    df.plot(ax=ax_1)

                if loading_st_data is not None:
                    p_st = np.arange(len(loading_st_data)).astype(float) / len(loading_st_data)
                    df = pd.DataFrame(data=loading_st_data,
                                      index=p_st,
                                      columns=[SimulationTypes.StochasticPowerFlow.value])
                    ax_1.set_title('Probability x < value', fontsize=14)
                    ax_1.set_ylabel('Loading [%]', fontsize=11)
                    df.plot(ax=ax_1)

                # power
                if len(power_data.keys()):
                    df = pd.DataFrame(data=power_data, index=x)
                    ax_2.set_title('Power', fontsize=14)
                    ax_2.set_ylabel('Power [MW]', fontsize=11)
                    df.plot(ax=ax_2)
                    ax_2.plot(x, api_object.rate_prof, c='gray', linestyle='dashed', linewidth=1)
                    ax_2.plot(x, -api_object.rate_prof, c='gray', linestyle='dashed', linewidth=1)

                if power_clustering_data is not None:
                    df = pd.DataFrame(data=power_clustering_data,
                                      index=x_cl,
                                      columns=[SimulationTypes.ClusteringTimeSeries_run.value])
                    ax_2.set_title('Power', fontsize=14)
                    ax_2.set_ylabel('Power [MW]', fontsize=11)
                    df.plot(ax=ax_2)

                plt.legend()
                fig.suptitle(api_object.name, fontsize=20)

                # plot the profiles
                plt.show()

    def plot_hvdc_branch(self, i, api_object):
        """

        :param api_object:
        :return:
        """
        fig = plt.figure(figsize=(12, 8))
        ax_1 = fig.add_subplot(211)
        # ax_2 = fig.add_subplot(212, sharex=ax_1)
        ax_2 = fig.add_subplot(212)

        # set time
        x = self.circuit.time_profile
        x_cl = x

        if x is not None:
            if len(x) > 0:

                p = np.arange(len(x)).astype(float) / len(x)

                # search available results
                power_data = dict()
                loading_data = dict()
                loading_st_data = None
                loading_clustering_data = None
                power_clustering_data = None

                for key, driver in self.results_dictionary.items():
                    if hasattr(driver, 'results'):
                        if driver.results is not None:
                            if key == SimulationTypes.TimeSeries_run:
                                power_data[key.value] = driver.results.hvdc_Pf[:, i]
                                loading_data[key.value] = np.sort(np.abs(driver.results.hvdc_loading[:, i] * 100.0))

                            elif key == SimulationTypes.ClusteringTimeSeries_run:
                                x_cl = x[driver.sampled_time_idx]
                                power_clustering_data = driver.results.hvdc_Pf[:, i]
                                loading_clustering_data = np.sort(np.abs(driver.results.hvdc_loading[:, i] * 100.0))

                            elif key == SimulationTypes.LinearAnalysis_TS_run:
                                power_data[key.value] = driver.results.hvdc_Pf[:, i]
                                loading_data[key.value] = np.sort(np.abs(driver.results.hvdc_loading[:, i] * 100.0))

                            elif key == SimulationTypes.OPFTimeSeries_run:
                                power_data[key.value] = driver.results.hvdc_Pf[:, i]
                                loading_data[key.value] = np.sort(np.abs(driver.results.hvdc_loading[:, i] * 100.0))

                # add the rating
                # power_data['Rates+'] = api_object.rate_prof
                # power_data['Rates-'] = -api_object.rate_prof

                # loading
                if len(loading_data.keys()):
                    df = pd.DataFrame(data=loading_data, index=p)
                    ax_1.set_title('Probability x < value', fontsize=14)
                    ax_1.set_ylabel('Loading [%]', fontsize=11)
                    df.plot(ax=ax_1)

                if loading_clustering_data is not None:
                    p_st = np.arange(len(loading_clustering_data)).astype(float) / len(loading_clustering_data)
                    df = pd.DataFrame(data=loading_clustering_data,
                                      index=p_st,
                                      columns=[SimulationTypes.ClusteringTimeSeries_run.value])
                    ax_1.set_title('Probability x < value', fontsize=14)
                    ax_1.set_ylabel('Loading [%]', fontsize=11)
                    df.plot(ax=ax_1)

                if loading_st_data is not None:
                    p_st = np.arange(len(loading_st_data)).astype(float) / len(loading_st_data)
                    df = pd.DataFrame(data=loading_st_data,
                                      index=p_st,
                                      columns=[SimulationTypes.StochasticPowerFlow.value])
                    ax_1.set_title('Probability x < value', fontsize=14)
                    ax_1.set_ylabel('Loading [%]', fontsize=11)
                    df.plot(ax=ax_1)

                # power
                if len(power_data.keys()):
                    df = pd.DataFrame(data=power_data, index=x)
                    ax_2.set_title('Power', fontsize=14)
                    ax_2.set_ylabel('Power [MW]', fontsize=11)
                    df.plot(ax=ax_2)
                    ax_2.plot(x, api_object.rate_prof, c='gray', linestyle='dashed', linewidth=1)
                    ax_2.plot(x, -api_object.rate_prof, c='gray', linestyle='dashed', linewidth=1)

                if power_clustering_data is not None:
                    df = pd.DataFrame(data=power_clustering_data,
                                      index=x_cl,
                                      columns=[SimulationTypes.ClusteringTimeSeries_run.value])
                    ax_2.set_title('Power', fontsize=14)
                    ax_2.set_ylabel('Power [MW]', fontsize=11)
                    df.plot(ax=ax_2)

                plt.legend()
                fig.suptitle(api_object.name, fontsize=20)

                # plot the profiles
                plt.show()

    def set_rate_to_profile(self, api_object):
        """

        :param api_object:
        """
        if api_object is not None:
            if api_object.rate_prof is not None:
                quit_msg = str(api_object.name) + \
                           "\nAre you sure that you want to overwrite the rates profile with the snapshot value?"
                reply = QMessageBox.question(self.parent_, 'Overwrite the profile', quit_msg,
                                             QMessageBox.Yes, QMessageBox.No)

                if reply == QMessageBox.Yes:
                    api_object.rate_prof *= 0
                    api_object.rate_prof += api_object.rate

    def set_active_status_to_profile(self, api_object, override_question=False):
        """

        :param api_object:
        :param override_question:
        :return:
        """
        if api_object is not None:
            if api_object.active_prof is not None:
                if not override_question:
                    quit_msg = str(api_object.name) + \
                               "\nAre you sure that you want to overwrite the active profile with the snapshot value?"
                    reply = QMessageBox.question(self.parent_, 'Overwrite the active profile', quit_msg,
                                                 QMessageBox.Yes, QMessageBox.No)
                    ok = reply == QMessageBox.Yes
                else:
                    ok = True

                if ok:
                    shape = api_object.active_prof.shape
                    if api_object.active:
                        api_object.active_prof = np.ones(shape, dtype=bool)
                    else:
                        api_object.active_prof = np.zeros(shape, dtype=bool)

    def mouseMoveEvent(self, event):
        """

        @param event:
        @return:
        """
        self.parent_.scene_mouse_move_event(event)

        # call the parent event
        super(DiagramScene, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        """

        @param event:
        @return:
        """
        self.parent_.scene_mouse_release_event(event)

        # call mouseReleaseEvent on "me" (continue with the rest of the actions)
        super(DiagramScene, self).mouseReleaseEvent(event)


class GridEditor(QSplitter):

    def __init__(self, circuit: MultiCircuit):
        """
        Creates the Diagram Editor
        Args:
            circuit: Circuit that is handling
        """
        QSplitter.__init__(self)

        # store a reference to the multi circuit instance
        self.circuit = circuit

        # nodes distance "explosion" factor
        self.expand_factor = 1.5

        # Widget layout and child widgets:
        self.horizontalLayout = QHBoxLayout(self)
        self.object_editor_table = QTableView(self)
        self.libraryBrowserView = QListView(self)
        self.libraryModel = LibraryModel(self)
        self.libraryModel.setColumnCount(1)

        # initialize library of items
        self.libItems = list()

        # add bus to the drag&drop
        bus_icon = QIcon()
        bus_icon.addPixmap(QPixmap(":/Icons/icons/bus_icon.svg"))
        item = QStandardItem(bus_icon, "Bus")
        item.setToolTip("Drag & drop this into the schematic")
        self.libItems.append(item)

        # add transformer3w to the drag&drop
        # t3w_icon = QIcon()
        # t3w_icon.addPixmap(QPixmap(":/Icons/icons/transformer3w.svg"))
        # item = QStandardItem(t3w_icon, "3W-Transformer")
        # item.setToolTip("Drag & drop this into the schematic")
        # self.libItems.append(item)

        for i in self.libItems:
            self.libraryModel.appendRow(i)

        # set the objects list
        self.object_types = [dev.device_type.value for dev in circuit.objects_with_profiles]

        self.catalogue_types = ['Wires', 'Overhead lines', 'Underground lines', 'Sequence lines', 'Transformers']

        # Actual libraryView object
        self.libraryBrowserView.setModel(self.libraryModel)
        self.libraryBrowserView.setViewMode(self.libraryBrowserView.ListMode)
        self.libraryBrowserView.setDragDropMode(self.libraryBrowserView.DragOnly)

        # create all the schematic objects and replace the existing ones
        self.diagramScene = DiagramScene(self, circuit)  # scene to add to the QGraphicsView
        self.diagramView = EditorGraphicsView(self.diagramScene, parent=self, editor=self)

        # create the grid name editor
        self.frame1 = QFrame()
        self.frame1_layout = QVBoxLayout()
        self.frame1_layout.setContentsMargins(0, 0, 0, 0)

        self.name_editor_frame = QFrame()
        self.name_layout = QHBoxLayout()
        self.name_layout.setContentsMargins(0, 0, 0, 0)

        self.name_label = QLineEdit()
        self.name_label.setText(str(self.circuit.name))
        self.name_label.setToolTip('Name of the model')
        self.name_layout.addWidget(self.name_label)
        self.name_editor_frame.setLayout(self.name_layout)

        self.frame1_layout.addWidget(self.name_editor_frame)
        self.frame1_layout.addWidget(self.libraryBrowserView)
        self.frame1.setLayout(self.frame1_layout)

        # Add the two objects into a layout
        splitter2 = QSplitter(self)
        splitter2.addWidget(self.frame1)
        splitter2.addWidget(self.object_editor_table)
        splitter2.setOrientation(Qt.Vertical)
        self.addWidget(splitter2)
        self.addWidget(self.diagramView)

        # factor 1:10
        splitter2.setStretchFactor(0, 1)
        splitter2.setStretchFactor(1, 5)

        self.started_branch = None

        self.setStretchFactor(0, 0)
        self.setStretchFactor(1, 2000)

    def start_connection(self, port: TerminalItem):
        """
        Start the branch creation
        @param port:
        @return:
        """
        self.started_branch = LineGraphicItem(fromPort=port, toPort=None, diagramScene=self.diagramScene)
        self.started_branch.bus_from = port.parent
        port.setZValue(0)
        port.process_callbacks(port.parent.pos() + port.pos())

    def scene_mouse_move_event(self, event):
        """

        @param event:
        @return:
        """
        if self.started_branch:
            pos = event.scenePos()
            self.started_branch.setEndPos(pos)

    def scene_mouse_release_event(self, event):
        """
        Finalize the branch creation if its drawing ends in a terminal
        @param event:
        @return:
        """
        # Clear or finnish the started connection:
        if self.started_branch:
            pos = event.scenePos()
            items = self.diagramScene.items(pos)  # get the item (the terminal) at the mouse position

            for item in items:
                if type(item) is TerminalItem:  # connect only to terminals
                    if item.parent is not self.started_branch.fromPort.parent:  # forbid connecting to itself

                        self.started_branch.setToPort(item)
                        item.hosting_connections.append(self.started_branch)
                        self.started_branch.bus_to = item.parent

                        if isinstance(self.started_branch.bus_from.api_object, Bus) and \
                           isinstance(self.started_branch.bus_to.api_object, Bus):

                            if self.started_branch.bus_from.api_object.is_dc != self.started_branch.bus_to.api_object.is_dc:
                                # different DC status -> VSC

                                name = 'VSC ' + str(len(self.circuit.vsc_devices) + 1)
                                obj = VSC(bus_from=self.started_branch.bus_from.api_object,
                                          bus_to=self.started_branch.bus_to.api_object,
                                          name=name)

                                obj.graphic_obj = VscGraphicItem(fromPort=self.started_branch.fromPort,
                                                                 toPort=self.started_branch.toPort,
                                                                 diagramScene=self.diagramScene,
                                                                 branch=obj)

                            elif self.started_branch.bus_from.api_object.is_dc and self.started_branch.bus_to.api_object.is_dc:
                                # both buses are DC

                                name = 'Dc line ' + str(len(self.circuit.dc_lines) + 1)
                                obj = DcLine(bus_from=self.started_branch.bus_from.api_object,
                                             bus_to=self.started_branch.bus_to.api_object,
                                             name=name)

                                obj.graphic_obj = DcLineGraphicItem(fromPort=self.started_branch.fromPort,
                                                                    toPort=self.started_branch.toPort,
                                                                    diagramScene=self.diagramScene,
                                                                    branch=obj)

                            else:
                                # Same DC status -> line / trafo
                                v1 = self.started_branch.bus_from.api_object.Vnom
                                v2 = self.started_branch.bus_to.api_object.Vnom

                                if abs(v1 - v2) > 1.0:
                                    name = 'Transformer ' + str(len(self.circuit.transformers2w) + 1)
                                    obj = Transformer2W(bus_from=self.started_branch.bus_from.api_object,
                                                        bus_to=self.started_branch.bus_to.api_object,
                                                        name=name)

                                    obj.graphic_obj = TransformerGraphicItem(fromPort=self.started_branch.fromPort,
                                                                             toPort=self.started_branch.toPort,
                                                                             diagramScene=self.diagramScene,
                                                                             branch=obj)

                                else:
                                    name = 'Line ' + str(len(self.circuit.lines) + 1)
                                    obj = Line(bus_from=self.started_branch.bus_from.api_object,
                                               bus_to=self.started_branch.bus_to.api_object,
                                               name=name)

                                    obj.graphic_obj = LineGraphicItem(fromPort=self.started_branch.fromPort,
                                                                      toPort=self.started_branch.toPort,
                                                                      diagramScene=self.diagramScene,
                                                                      branch=obj)

                            # add the new object to the circuit
                            self.circuit.add_branch(obj)

                            # update the connection placement
                            obj.graphic_obj.fromPort.update()
                            obj.graphic_obj.toPort.update()

                            # set the connection placement
                            obj.graphic_obj.setZValue(-1)

                        elif isinstance(self.started_branch.bus_from.api_object, Transformer3W):

                            obj = self.started_branch.bus_from.api_object

                            obj.graphic_obj = LineGraphicItem(fromPort=self.started_branch.fromPort,
                                                              toPort=self.started_branch.toPort,
                                                              diagramScene=self.diagramScene,
                                                              branch=None)

                            # update the connection placement
                            obj.graphic_obj.fromPort.update()
                            obj.graphic_obj.toPort.update()

                            # set the connection placement
                            obj.graphic_obj.setZValue(-1)

                            print('Hosted tr3w connection FROM')

                        elif isinstance(self.started_branch.bus_to.api_object, Transformer3W):

                            obj = self.started_branch.bus_to.api_object

                            obj.graphic_obj = LineGraphicItem(fromPort=self.started_branch.fromPort,
                                                              toPort=self.started_branch.toPort,
                                                              diagramScene=self.diagramScene,
                                                              branch=None)

                            # update the connection placement
                            obj.graphic_obj.fromPort.update()
                            obj.graphic_obj.toPort.update()

                            # set the connection placement
                            obj.graphic_obj.setZValue(-1)

                            print('Hosted tr3w connection TO')

                        else:
                            print('unknown connection')

            # if self.started_branch.toPort is None:
            self.started_branch.remove_widget()

        # release this pointer
        self.started_branch = None

    def bigger_nodes(self):
        """
        Expand the grid
        @return:
        """
        min_x = sys.maxsize
        min_y = sys.maxsize
        max_x = -sys.maxsize
        max_y = -sys.maxsize

        if len(self.diagramScene.selectedItems()) > 0:

            # expand selection
            for item in self.diagramScene.selectedItems():
                if type(item) is BusGraphicItem:
                    x = item.pos().x() * self.expand_factor
                    y = item.pos().y() * self.expand_factor
                    item.setPos(QPointF(x, y))

                    # apply changes to the API objects
                    if item.api_object is not None:
                        item.api_object.x = x
                        item.api_object.y = y

                    max_x = max(max_x, x)
                    min_x = min(min_x, x)
                    max_y = max(max_y, y)
                    min_y = min(min_y, y)

        else:

            # expand all
            for item in self.diagramScene.items():
                if type(item) is BusGraphicItem:
                    x = item.pos().x() * self.expand_factor
                    y = item.pos().y() * self.expand_factor
                    item.setPos(QPointF(x, y))

                    max_x = max(max_x, x)
                    min_x = min(min_x, x)
                    max_y = max(max_y, y)
                    min_y = min(min_y, y)

                    # apply changes to the API objects
                    if item.api_object is not None:
                        item.api_object.x = x
                        item.api_object.y = y

        # set the limits of the view
        self.set_limits(min_x, max_x, min_y, max_y)

    def smaller_nodes(self):
        """
        Contract the grid
        @return:
        """
        min_x = sys.maxsize
        min_y = sys.maxsize
        max_x = -sys.maxsize
        max_y = -sys.maxsize

        if len(self.diagramScene.selectedItems()) > 0:

            # shrink selection only
            for item in self.diagramScene.selectedItems():
                if type(item) is BusGraphicItem:
                    x = item.pos().x() / self.expand_factor
                    y = item.pos().y() / self.expand_factor
                    item.setPos(QPointF(x, y))

                    # apply changes to the API objects
                    if item.api_object is not None:
                        item.api_object.x = x
                        item.api_object.y = y

                    max_x = max(max_x, x)
                    min_x = min(min_x, x)
                    max_y = max(max_y, y)
                    min_y = min(min_y, y)
        else:

            # shrink all
            for item in self.diagramScene.items():
                if type(item) is BusGraphicItem:
                    x = item.pos().x() / self.expand_factor
                    y = item.pos().y() / self.expand_factor
                    item.setPos(QPointF(x, y))

                    # apply changes to the API objects
                    if item.api_object is not None:
                        item.api_object.x = x
                        item.api_object.y = y

                    max_x = max(max_x, x)
                    min_x = min(min_x, x)
                    max_y = max(max_y, y)
                    min_y = min(min_y, y)

        # set the limits of the view
        self.set_limits(min_x, max_x, min_y, max_y)

    def set_limits(self, min_x, max_x, min_y, max_y, margin_factor=0.1):
        """
        Set the picture limits
        :param min_x: Minimum x value of the buses location
        :param max_x: Maximum x value of the buses location
        :param min_y: Minimum y value of the buses location
        :param max_y: Maximum y value of the buses location
        :param margin_factor: factor of separation between the buses
        """
        dx = max_x - min_x
        dy = max_y - min_y
        mx = margin_factor * dx
        my = margin_factor * dy
        h = dy + 2 * my + 80
        w = dx + 2 * mx + 80
        self.diagramScene.setSceneRect(min_x - mx, min_y - my, w, h)

    def center_nodes(self):
        """
        Center the view in the nodes
        @return: Nothing
        """
        self.diagramView.fitInView(self.diagramScene.sceneRect(), Qt.KeepAspectRatio)
        self.diagramView.scale(1.0, 1.0)

    def auto_layout(self):
        """
        Automatic layout of the nodes
        """

        if self.circuit.graph is None:
            self.circuit.build_graph()

        pos = nx.spectral_layout(self.circuit.graph, scale=2, weight='weight')

        pos = nx.fruchterman_reingold_layout(self.circuit.graph, dim=2,
                                             k=None, pos=pos, fixed=None, iterations=500,
                                             weight='weight', scale=20.0, center=None)

        # assign the positions to the graphical objects of the nodes
        for i, bus in enumerate(self.circuit.buses):
            try:
                x, y = pos[i] * 500
                bus.graphic_obj.setPos(QPoint(x, y))

                # apply changes to the API objects
                bus.x = x
                bus.y = y

            except KeyError as ex:
                warn('auto_layout: Node ' + str(i) + ' not in the graph!!!! \n' + str(ex))

        self.center_nodes()

    def export(self, filename, w=1920, h=1080):
        """
        Save the grid to a png file
        """

        name, extension = os.path.splitext(filename.lower())

        if extension == '.png':
            image = QImage(w, h, QImage.Format_ARGB32_Premultiplied)
            image.fill(Qt.transparent)
            painter = QPainter(image)
            painter.setRenderHint(QPainter.Antialiasing)
            self.diagramScene.render(painter)
            image.save(filename)
            painter.end()

        elif extension == '.svg':
            svg_gen = QSvgGenerator()
            svg_gen.setFileName(filename)
            svg_gen.setSize(QSize(w, h))
            svg_gen.setViewBox(QRect(0, 0, w, h))
            svg_gen.setTitle("Electrical grid schematic")
            svg_gen.setDescription("An SVG drawing created by GridCal")

            painter = QPainter(svg_gen)
            self.diagramScene.render(painter)
            painter.end()
        else:
            raise Exception('Extension ' + str(extension) + ' not supported :(')

    def add_line(self, branch: Line):
        """
        Add branch to the schematic
        :param branch: Branch object
        """
        terminal_from = branch.bus_from.graphic_obj.terminal
        terminal_to = branch.bus_to.graphic_obj.terminal
        graphic_obj = LineGraphicItem(terminal_from, terminal_to, self.diagramScene, branch=branch)
        graphic_obj.diagramScene.circuit = self.circuit  # add pointer to the circuit
        terminal_from.hosting_connections.append(graphic_obj)
        terminal_to.hosting_connections.append(graphic_obj)
        graphic_obj.redraw()
        branch.graphic_obj = graphic_obj

    def add_dc_line(self, branch: DcLine):
        """
        Add branch to the schematic
        :param branch: Branch object
        """
        terminal_from = branch.bus_from.graphic_obj.terminal
        terminal_to = branch.bus_to.graphic_obj.terminal
        graphic_obj = DcLineGraphicItem(terminal_from, terminal_to, self.diagramScene, branch=branch)
        graphic_obj.diagramScene.grid = self.circuit  # add pointer to the circuit
        terminal_from.hosting_connections.append(graphic_obj)
        terminal_to.hosting_connections.append(graphic_obj)
        graphic_obj.redraw()
        branch.graphic_obj = graphic_obj

    def add_transformer(self, branch: Transformer2W):
        """
        Add branch to the schematic
        :param branch: Branch object
        """
        terminal_from = branch.bus_from.graphic_obj.terminal
        terminal_to = branch.bus_to.graphic_obj.terminal
        graphic_obj = TransformerGraphicItem(terminal_from, terminal_to, self.diagramScene, branch=branch)
        graphic_obj.diagramScene.circuit = self.circuit  # add pointer to the circuit
        terminal_from.hosting_connections.append(graphic_obj)
        terminal_to.hosting_connections.append(graphic_obj)
        graphic_obj.redraw()
        branch.graphic_obj = graphic_obj

    def add_api_bus(self, bus: Bus, explode_factor=1.0):
        """
        Add API bus to the diagram
        :param bus: Bus instance
        :param explode_factor: explode factor
        """
        # add the graphic object to the diagram view
        graphic_obj = self.diagramView.add_bus(bus=bus, explode_factor=explode_factor)

        # add circuit pointer to the bus graphic element
        graphic_obj.diagramScene.circuit = self.circuit  # add pointer to the circuit

        # create the bus children
        graphic_obj.create_children_icons()

        # arrange the children
        graphic_obj.arrange_children()

        return graphic_obj

    def add_api_line(self, branch: Line):
        """
        add API branch to the Scene
        :param branch: Branch instance
        """
        terminal_from = branch.bus_from.graphic_obj.terminal
        terminal_to = branch.bus_to.graphic_obj.terminal

        graphic_obj = LineGraphicItem(terminal_from, terminal_to, self.diagramScene, branch=branch)

        graphic_obj.diagramScene.circuit = self.circuit  # add pointer to the circuit
        terminal_from.hosting_connections.append(graphic_obj)
        terminal_to.hosting_connections.append(graphic_obj)
        graphic_obj.redraw()

        return graphic_obj

    def add_api_dc_line(self, branch: DcLine):
        """
        add API branch to the Scene
        :param branch: Branch instance
        """
        terminal_from = branch.bus_from.graphic_obj.terminal
        terminal_to = branch.bus_to.graphic_obj.terminal

        graphic_obj = DcLineGraphicItem(terminal_from, terminal_to, self.diagramScene, branch=branch)

        graphic_obj.diagramScene.grid = self.circuit  # add pointer to the circuit
        terminal_from.hosting_connections.append(graphic_obj)
        terminal_to.hosting_connections.append(graphic_obj)
        graphic_obj.redraw()

        return graphic_obj

    def add_api_hvdc(self, branch: HvdcLine):
        """
        add API branch to the Scene
        :param branch: Branch instance
        """
        terminal_from = branch.bus_from.graphic_obj.terminal
        terminal_to = branch.bus_to.graphic_obj.terminal

        graphic_obj = HvdcGraphicItem(terminal_from, terminal_to, self.diagramScene, branch=branch)

        graphic_obj.diagramScene.grid = self.circuit  # add pointer to the circuit
        terminal_from.hosting_connections.append(graphic_obj)
        terminal_to.hosting_connections.append(graphic_obj)
        graphic_obj.redraw()

        return graphic_obj

    def add_api_vsc(self, branch: VSC):
        """
        add API branch to the Scene
        :param branch: Branch instance
        """
        terminal_from = branch.bus_from.graphic_obj.terminal
        terminal_to = branch.bus_to.graphic_obj.terminal

        graphic_obj = VscGraphicItem(terminal_from, terminal_to, self.diagramScene, branch=branch)

        graphic_obj.diagramScene.grid = self.circuit  # add pointer to the circuit
        terminal_from.hosting_connections.append(graphic_obj)
        terminal_to.hosting_connections.append(graphic_obj)
        graphic_obj.redraw()

        return graphic_obj

    def add_api_upfc(self, branch: UPFC):
        """
        add API branch to the Scene
        :param branch: Branch instance
        """
        terminal_from = branch.bus_from.graphic_obj.terminal
        terminal_to = branch.bus_to.graphic_obj.terminal

        graphic_obj = UpfcGraphicItem(terminal_from, terminal_to, self.diagramScene, branch=branch)

        graphic_obj.diagramScene.grid = self.circuit  # add pointer to the circuit
        terminal_from.hosting_connections.append(graphic_obj)
        terminal_to.hosting_connections.append(graphic_obj)
        graphic_obj.redraw()

        return graphic_obj

    def add_api_transformer(self, branch: Transformer2W):
        """
        add API branch to the Scene
        :param branch: Branch instance
        """
        terminal_from = branch.bus_from.graphic_obj.terminal
        terminal_to = branch.bus_to.graphic_obj.terminal

        graphic_obj = TransformerGraphicItem(terminal_from, terminal_to, self.diagramScene, branch=branch)

        graphic_obj.diagramScene.circuit = self.circuit  # add pointer to the circuit
        terminal_from.hosting_connections.append(graphic_obj)
        terminal_to.hosting_connections.append(graphic_obj)
        graphic_obj.redraw()

        return graphic_obj

    def convert_line_to_hvdc(self, line: Line):
        """
        Convert a line to HVDC, this is the GUI way to create HVDC objects
        :param line: Line instance
        :return: Nothing
        """
        hvdc = self.circuit.convert_line_to_hvdc(line)

        # add device to the schematic
        hvdc.graphic_obj = self.add_api_hvdc(hvdc)

        # update position
        hvdc.graphic_obj.fromPort.update()
        hvdc.graphic_obj.toPort.update()

        # delete from the schematic
        self.diagramScene.removeItem(line.graphic_obj)

    def convert_line_to_transformer(self, line: Line):
        """
        Convert a line to Transformer
        :param line: Line instance
        :return: Nothing
        """
        transformer = self.circuit.convert_line_to_transformer(line)

        # add device to the schematic
        transformer.graphic_obj = self.add_api_transformer(transformer)

        # update position
        transformer.graphic_obj.fromPort.update()
        transformer.graphic_obj.toPort.update()

        # delete from the schematic
        self.diagramScene.removeItem(line.graphic_obj)

    def convert_line_to_vsc(self, line: Line):
        """
        Convert a line to voltage source converter
        :param line: Line instance
        :return: Nothing
        """
        vsc = self.circuit.convert_line_to_vsc(line)

        # add device to the schematic
        vsc.graphic_obj = self.add_api_vsc(vsc)

        # update position
        vsc.graphic_obj.fromPort.update()
        vsc.graphic_obj.toPort.update()

        # delete from the schematic
        self.diagramScene.removeItem(line.graphic_obj)

    def convert_line_to_upfc(self, line: Line):
        """
        Convert a line to voltage source converter
        :param line: Line instance
        :return: Nothing
        """
        upfc = self.circuit.convert_line_to_upfc(line)

        # add device to the schematic
        upfc.graphic_obj = self.add_api_upfc(upfc)

        # update position
        upfc.graphic_obj.fromPort.update()
        upfc.graphic_obj.toPort.update()

        # delete from the schematic
        self.diagramScene.removeItem(line.graphic_obj)

    def add_elements_to_schematic(self, buses, lines, dc_lines, transformers2w, hvdc_lines, vsc_devices,
                                  upfc_devices, explode_factor=1.0, prog_func=None, text_func=None):
        """
        Add a elements to the schematic scene
        :param buses: list of Bus objects
        :param lines: list of Line objects
        :param dc_lines: list of DcLine objects
        :param transformers2w: list of Transformer Objects
        :param hvdc_lines: list of HvdcLine objects
        :param vsc_devices: list Vsc objects
        :param upfc_devices: List of UPFC devices
        :param explode_factor: factor of "explosion": Separation of the nodes factor
        :param prog_func: progress report function
        :param text_func: Text report function
        """
        # first create the buses
        if text_func is not None:
            text_func('Creating schematic buses')

        nn = len(buses)
        for i, bus in enumerate(buses):

            if prog_func is not None:
                prog_func((i+1) / nn * 100.0)

            bus.graphic_obj = self.add_api_bus(bus, explode_factor)

        # --------------------------------------------------------------------------------------------------------------
        if text_func is not None:
            text_func('Creating schematic line devices')

        nn = len(lines)
        for i, branch in enumerate(lines):

            if prog_func is not None:
                prog_func((i+1) / nn * 100.0)

            branch.graphic_obj = self.add_api_line(branch)

        # --------------------------------------------------------------------------------------------------------------
        if text_func is not None:
            text_func('Creating schematic line devices')

        nn = len(dc_lines)
        for i, branch in enumerate(dc_lines):

            if prog_func is not None:
                prog_func((i + 1) / nn * 100.0)

            branch.graphic_obj = self.add_api_dc_line(branch)

        # --------------------------------------------------------------------------------------------------------------
        if text_func is not None:
            text_func('Creating schematic transformer devices')

        nn = len(transformers2w)
        for i, branch in enumerate(transformers2w):

            if prog_func is not None:
                prog_func((i + 1) / nn * 100.0)

            branch.graphic_obj = self.add_api_transformer(branch)

        # --------------------------------------------------------------------------------------------------------------
        if text_func is not None:
            text_func('Creating schematic HVDC devices')

        nn = len(hvdc_lines)
        for i, branch in enumerate(hvdc_lines):

            if prog_func is not None:
                prog_func((i + 1) / nn * 100.0)

            branch.graphic_obj = self.add_api_hvdc(branch)

        # --------------------------------------------------------------------------------------------------------------
        if text_func is not None:
            text_func('Creating schematic VSC devices')

        nn = len(vsc_devices)
        for i, branch in enumerate(vsc_devices):

            if prog_func is not None:
                prog_func((i + 1) / nn * 100.0)

            branch.graphic_obj = self.add_api_vsc(branch)

        # --------------------------------------------------------------------------------------------------------------
        if text_func is not None:
            text_func('Creating schematic UPFC devices')

        nn = len(upfc_devices)
        for i, branch in enumerate(upfc_devices):

            if prog_func is not None:
                prog_func((i + 1) / nn * 100.0)

            branch.graphic_obj = self.add_api_upfc(branch)

    def add_circuit_to_schematic(self, circuit: "MultiCircuit", explode_factor=1.0, prog_func=None, text_func=None):
        """
        Add a complete circuit to the schematic scene
        :param circuit: MultiCircuit instance
        :param explode_factor: factor of "explosion": Separation of the nodes factor
        :param prog_func: progress report function
        :param text_func: Text report function
        """

        # reset zoom level, otherwise the newly loaded grids appear with a much wider spacing
        self.diagramView.resetTransform()

        self.add_elements_to_schematic(buses=circuit.buses,
                                       lines=circuit.lines,
                                       dc_lines=circuit.dc_lines,
                                       transformers2w=circuit.transformers2w,
                                       hvdc_lines=circuit.hvdc_lines,
                                       vsc_devices=circuit.vsc_devices,
                                       upfc_devices=circuit.upfc_devices,
                                       explode_factor=explode_factor,
                                       prog_func=prog_func,
                                       text_func=text_func)

    def align_schematic(self, buses: List[Bus] = []):
        """
        Align the scene view to the content
        """
        # figure limits
        min_x = sys.maxsize
        min_y = sys.maxsize
        max_x = -sys.maxsize
        max_y = -sys.maxsize

        if len(buses):
            lst = buses
        else:
            lst = self.circuit.buses

        # Align lines
        for bus in lst:
            bus.graphic_obj.arrange_children()
            # get the item position
            x = bus.graphic_obj.pos().x()
            y = bus.graphic_obj.pos().y()

            # compute the boundaries of the grid
            max_x = max(max_x, x)
            min_x = min(min_x, x)
            max_y = max(max_y, y)
            min_y = min(min_y, y)

        # set the figure limits
        self.set_limits(min_x, max_x, min_y, max_y)

        #  center the view
        self.center_nodes()

    def clear(self):
        self.diagramView.scene_.clear()
        self.name_label.setText("")

    def schematic_from_api(self, explode_factor=1.0, prog_func=None, text_func=None):
        """
        Generate schematic from the API
        :param explode_factor: factor to separate the nodes
        :param prog_func: progress report function
        :param text_func: Text report function
        :return: Nothing
        """
        # clear all
        self.clear()

        # add to schematic
        self.add_circuit_to_schematic(self.circuit,
                                      explode_factor=explode_factor,
                                      prog_func=prog_func,
                                      text_func=text_func)

        if text_func is not None:
            text_func('Aligning schematic...')

        self.align_schematic()



if __name__ == '__main__':
    app = QApplication(sys.argv)
    app.setStyle('Fusion')  # ['Breeze', 'Oxygen', 'QtCurve', 'Windows', 'Fusion']
    circuit = MultiCircuit()
    window = GridEditor(circuit)
    window.show()
    sys.exit(app.exec_())


# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import time

from scipy.sparse.linalg import spsolve

from GridCal.Engine.Core.admittance_matrices import compile_y_acdc
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.common_functions import *
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.acdc_jacobian import fubm_jacobian, AcDcSolSlicer


def LM_ACDC(nc: "SnapshotData", Vbus, S0, I0, Y0,
            tolerance=1e-6, max_iter=4, verbose=False) -> NumericPowerFlowResults:
    """
    Solves the power flow problem by the Levenberg-Marquardt power flow algorithm.
    It is usually better than Newton-Raphson, but it takes an order of magnitude more time to converge.

    :param nc: SnapshotData instance
    :param tolerance: maximum error allowed
    :param max_iter: maximum number of iterations
    :return:
    """
    start = time.time()

    # initialize the variables
    nb = nc.nbus
    nl = nc.nbr
    V = Vbus

    Va = np.angle(V)
    Vm = np.abs(V)

    Sbus = S0 + I0 * Vm + Y0 * np.power(Vm, 2)  # compute the ZIP power injection

    Vmfset = nc.branch_data.vf_set[:, 0]
    m = nc.branch_data.m[:, 0].copy()
    theta = nc.branch_data.theta[:, 0].copy()
    Beq = nc.branch_data.Beq[:, 0].copy()
    Gsw = nc.branch_data.G0[:, 0]
    Pfset = nc.branch_data.Pfset[:, 0] / nc.Sbase
    Qfset = nc.branch_data.Qfset[:, 0] / nc.Sbase
    Qtset = nc.branch_data.Qfset[:, 0] / nc.Sbase
    Kdp = nc.branch_data.Kdp
    k2 = nc.branch_data.k
    Cf = nc.Cf
    Ct = nc.Ct
    F = nc.F
    T = nc.T
    Ys = 1.0 / (nc.branch_data.R + 1j * nc.branch_data.X)
    Bc = nc.branch_data.B
    pq = nc.pq.copy().astype(int)
    pvpq_orig = np.r_[nc.pv, pq].astype(int)
    pvpq_orig.sort()

    # the elements of PQ that exist in the control indices Ivf and Ivt must be passed from the PQ to the PV list
    # otherwise those variables would be in two sets of equations
    i_ctrl_v = np.unique(np.r_[nc.VfBeqbus, nc.Vtmabus])
    for val in pq:
        if val in i_ctrl_v:
            pq = pq[pq != val]

    # compose the new pvpq indices à la NR
    pv = np.unique(np.r_[i_ctrl_v, nc.pv]).astype(int)
    pv.sort()
    pvpq = np.r_[pv, pq].astype(int)
    npv = len(pv)
    npq = len(pq)

    if (npq + npv) > 0:
        # --------------------------------------------------------------------------
        # variables dimensions in Jacobian
        sol_slicer = AcDcSolSlicer(npq, npv,
                                   len(nc.VfBeqbus),
                                   len(nc.Vtmabus),
                                   len(nc.iPfsh),
                                   len(nc.iQfma),
                                   len(nc.iBeqz),
                                   len(nc.iQtma),
                                   len(nc.iPfdp))
        # -------------------------------------------------------------------------
        # compute initial admittances
        Ybus, Yf, Yt, tap = compile_y_acdc(Cf=Cf, Ct=Ct,
                                           C_bus_shunt=nc.shunt_data.C_bus_shunt,
                                           shunt_admittance=nc.shunt_data.shunt_admittance[:, 0],
                                           shunt_active=nc.shunt_data.shunt_active[:, 0],
                                           ys=Ys,
                                           B=Bc,
                                           Sbase=nc.Sbase,
                                           m=m, theta=theta, Beq=Beq, Gsw=Gsw,
                                           mf=nc.branch_data.tap_f,
                                           mt=nc.branch_data.tap_t)

        #  compute branch power Sf
        If = Yf * V  # complex current injected at "from" bus, Yf(br, :) * V; For in-service branches
        It = Yt * V  # complex current injected at "to"   bus, Yt(br, :) * V; For in-service branches
        Sf = V[F] * np.conj(If)  # complex power injected at "from" bus
        St = V[T] * np.conj(It)  # complex power injected at "to"   bus

        # compute converter losses
        Gsw = compute_converter_losses(V=V, It=It, F=F,
                                       alpha1=nc.branch_data.alpha1,
                                       alpha2=nc.branch_data.alpha2,
                                       alpha3=nc.branch_data.alpha3,
                                       iVscL=nc.iVscL)

        # compute total mismatch
        Scalc = compute_power(Ybus, V)
        dz = compute_acdc_fx(Vm=Vm,
                             Sbus=Sbus,
                             Scalc=Scalc,
                             Sf=Sf,
                             St=St,
                             Pfset=Pfset,
                             Qfset=Qfset,
                             Qtset=Qtset,
                             Vmfset=Vmfset,
                             Kdp=Kdp,
                             F=F,
                             pvpq=pvpq,
                             pq=pq,
                             iPfsh=nc.iPfsh,
                             iQfma=nc.iQfma,
                             iBeqz=nc.iBeqz,
                             iQtma=nc.iQtma,
                             iPfdp=nc.iPfdp,
                             VfBeqbus=nc.VfBeqbus,
                             Vtmabus=nc.Vtmabus)

        norm_f = np.max(np.abs(dz))

        update_jacobian = True
        converged = norm_f < tolerance
        iter_ = 0
        nu = 2.0
        lbmda = 0
        f_prev = 1e9  # very large number

        # generate lookup pvpq -> index pvpq (used in createJ)
        pvpq_lookup = np.zeros(Ybus.shape[0], dtype=int)
        pvpq_lookup[pvpq] = np.arange(len(pvpq))

        while not converged and iter_ < max_iter:

            # evaluate Jacobian
            if update_jacobian:
                H = fubm_jacobian(nb, nl, nc.iPfsh, nc.iPfdp, nc.iQfma, nc.iQtma, nc.iVtma, nc.iBeqz, nc.iBeqv,
                                  nc.VfBeqbus, nc.Vtmabus,
                                  F, T, Ys, k2, tap, m, Bc, Beq, Kdp, V, Ybus, Yf, Yt, Cf, Ct, pvpq, pq)

                if iter_ == 0:
                    # compute this identity only once
                    Idn = sp.diags(np.ones(H.shape[0]))  # csc_matrix identity

                # system matrix
                # H1 = H^t
                H1 = H.transpose()

                # H2 = H1·H
                H2 = H1.dot(H)

                # set first value of lmbda
                if iter_ == 0:
                    lbmda = 1e-3 * H2.diagonal().max()

            # compute system matrix A = H^T·H - lambda·I
            A = H2 + lbmda * Idn

            # right hand side
            # H^t·dz
            rhs = H1.dot(dz)

            # Solve the increment
            dx = spsolve(A, rhs)

            # objective function to minimize
            f = 0.5 * dz.dot(dz)

            # decision function
            val = dx.dot(lbmda * dx + rhs)
            if val > 0.0:
                rho = (f_prev - f) / (0.5 * val)
            else:
                rho = -1.0

            # lambda update
            if rho >= 0:
                update_jacobian = True
                lbmda *= max([1.0 / 3.0, 1 - (2 * rho - 1) ** 3])
                nu = 2.0

                # split the solution
                dVa, dVm, dBeq_v, dma_Vt, dtheta_Pf, dma_Qf, dBeq_z, dma_Qt, dtheta_Pd = sol_slicer.split(dx)

                # assign the new values
                Va[pvpq] -= dVa
                Vm[pq] -= dVm
                theta[nc.iPfsh] -= dtheta_Pf
                theta[nc.iPfdp] -= dtheta_Pd
                m[nc.iQfma] -= dma_Qf
                m[nc.iQtma] -= dma_Qt
                m[nc.iVtma] -= dma_Vt
                Beq[nc.iBeqz] -= dBeq_z
                Beq[nc.iBeqv] -= dBeq_v
                V = polar_to_rect(Vm, Va)
                Sbus = S0 + I0 * Vm + Y0 * np.power(Vm, 2)  # compute the ZIP power injection

            else:
                update_jacobian = False
                lbmda *= nu
                nu *= 2.0

            # compute initial admittances
            Ybus, Yf, Yt, tap = compile_y_acdc(Cf=Cf, Ct=Ct,
                                               C_bus_shunt=nc.shunt_data.C_bus_shunt,
                                               shunt_admittance=nc.shunt_data.shunt_admittance[:, 0],
                                               shunt_active=nc.shunt_data.shunt_active[:, 0],
                                               ys=Ys,
                                               B=Bc,
                                               Sbase=nc.Sbase,
                                               m=m, theta=theta, Beq=Beq, Gsw=Gsw,
                                               mf=nc.branch_data.tap_f,
                                               mt=nc.branch_data.tap_t)

            #  compute branch power Sf
            If = Yf * V  # complex current injected at "from" bus, Yf(br, :) * V; For in-service branches
            It = Yt * V  # complex current injected at "to"   bus, Yt(br, :) * V; For in-service branches
            Sf = V[F] * np.conj(If)  # complex power injected at "from" bus
            St = V[T] * np.conj(It)  # complex power injected at "to"   bus

            # compute converter losses
            Gsw = compute_converter_losses(V=V, It=It, F=F,
                                           alpha1=nc.branch_data.alpha1,
                                           alpha2=nc.branch_data.alpha2,
                                           alpha3=nc.branch_data.alpha3,
                                           iVscL=nc.iVscL)

            # check convergence
            Scalc = compute_power(Ybus, V)
            dz = compute_acdc_fx(Vm=Vm,
                                 Sbus=Sbus,
                                 Scalc=Scalc,
                                 Sf=Sf,
                                 St=St,
                                 Pfset=Pfset,
                                 Qfset=Qfset,
                                 Qtset=Qtset,
                                 Vmfset=Vmfset,
                                 Kdp=Kdp,
                                 F=F,
                                 pvpq=pvpq,
                                 pq=pq,
                                 iPfsh=nc.iPfsh,
                                 iQfma=nc.iQfma,
                                 iBeqz=nc.iBeqz,
                                 iQtma=nc.iQtma,
                                 iPfdp=nc.iPfdp,
                                 VfBeqbus=nc.VfBeqbus,
                                 Vtmabus=nc.Vtmabus)

            norm_f = np.max(np.abs(dz))
            converged = norm_f < tolerance
            f_prev = f

            if verbose:
                print('dx:', dx)
                print('Va:', Va)
                print('Vm:', Vm)
                print('theta:', theta)
                print('ma:', m)
                print('Beq:', Beq)
                print('norm_f:', norm_f)

            # update iteration counter
            iter_ += 1
    else:
        norm_f = 0
        converged = True
        Scalc = S0 + I0 * Vm + Y0 * np.power(Vm, 2)  # compute the ZIP power injection
        iter_ = 0
        Ybus = None
        Yf = None
        Yt = None

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(V, converged, norm_f, Scalc, m, theta, Beq, Ybus, Yf, Yt, iter_, elapsed)




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from PySide2 import QtWidgets


def info_msg(text, title="Information"):
    """
    Message box
    :param text: Text to display
    :param title: Name of the window
    """
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Information)
    msg.setText(text)
    msg.setWindowTitle(title)
    msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
    return msg.exec_()


def warning_msg(text, title="Warning"):
    """
    Message box
    :param text: Text to display
    :param title: Name of the window
    """
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Warning)
    msg.setText(text)
    msg.setWindowTitle(title)
    msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
    return msg.exec_()


def error_msg(text, title="Error"):
    """
    Message box
    :param text: Text to display
    :param title: Name of the window
    """
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Critical)
    msg.setText(text)
    msg.setWindowTitle(title)
    msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
    return msg.exec_()


def yes_no_question(text, title='Question'):
    """
    Question message
    :param text:
    :param title:
    :return: True / False
    """
    buttonReply = QtWidgets.QMessageBox.question(None,
                                                 title,
                                                 text,
                                                 QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                                                 QtWidgets.QMessageBox.No)
    return buttonReply == QtWidgets.QMessageBox.Yes



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from PySide2.QtWidgets import *
import sys


class SampleObject:
    def __init__(self, name='o'):
        self.name = name

    def __str__(self):
        return self.name


class ObjectSelectWindow(QDialog):
    def __init__(self, title, object_list, parent=None):
        QDialog.__init__(self, parent)

        self.setWindowTitle(title)

        layout = QGridLayout()
        self.setLayout(layout)
        self.object_list = object_list
        self.list_widget = QListWidget()

        self.selected_object = None

        for i, obj in enumerate(object_list):
            self.list_widget.insertItem(i, obj.name)

        # self.list_widget.clicked.connect(self.clicked)
        self.list_widget.itemDoubleClicked.connect(self.dbl_clicked)
        layout.addWidget(self.list_widget)

    def dbl_clicked(self, qmodelindex):
        """
        Double clicked selection
        :param qmodelindex:
        :return:
        """
        idx = self.list_widget.currentIndex().row()
        if idx > -1:
            print(self.object_list[idx])
            self.selected_object = self.object_list[idx]
            self.close()


if __name__ == '__main__':
    app = QApplication(sys.argv)
    lst = [SampleObject(name='Object {}'.format(i)) for i in range(5)]
    screen = ObjectSelectWindow('Cosas', lst)
    screen.show()
    sys.exit(app.exec_())



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import numpy as np

import GridCal.Engine.basic_structures as bs
from GridCal.Engine.Simulations.StateEstimation.state_estimation import solve_se_lm
from GridCal.Engine.Simulations.PowerFlow.power_flow_worker import PowerFlowResults, power_flow_post_process
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Core.snapshot_pf_data import compile_snapshot_circuit
from GridCal.Engine.Devices.measurement import MeasurementType
from GridCal.Engine.Simulations.driver_template import DriverTemplate


class StateEstimationInput:

    def __init__(self):
        """
        State estimation inputs constructor
        """

        # Node active power measurements vector of pointers
        self.p_inj =list()

        # Node  reactive power measurements vector of pointers
        self.q_inj = list()

        # Branch active power measurements vector of pointers
        self.p_flow = list()

        # Branch reactive power measurements vector of pointers
        self.q_flow = list()

        # Branch current module measurements vector of pointers
        self.i_flow = list()

        # Node voltage module measurements vector of pointers
        self.vm_m = list()

        # nodes without power injection measurements
        self.p_inj_idx = list()

        # branches without power measurements
        self.p_flow_idx = list()

        # nodes without reactive power injection measurements
        self.q_inj_idx = list()

        # branches without reactive power measurements
        self.q_flow_idx = list()

        # branches without current measurements
        self.i_flow_idx = list()

        # nodes without voltage module measurements
        self.vm_m_idx = list()

    def clear(self):
        """
        Clear
        """
        self.p_inj.clear()
        self.p_flow.clear()
        self.q_inj.clear()
        self.q_flow.clear()
        self.i_flow.clear()
        self.vm_m.clear()

        self.p_inj_idx.clear()
        self.p_flow_idx.clear()
        self.q_inj_idx.clear()
        self.q_flow_idx.clear()
        self.i_flow_idx.clear()
        self.vm_m_idx.clear()

    def consolidate(self):
        """
        consolidate the measurements into "measurements" and "sigma"
        :return: measurements, sigma
        """

        nz = len(self.p_inj) + len(self.p_flow) + len(self.q_inj) + len(self.q_flow) + len(self.i_flow) + len(self.vm_m)

        magnitudes = np.zeros(nz)
        sigma = np.zeros(nz)

        # go through the measurements in order and form the vectors
        k = 0
        for m in self.p_flow + self.p_inj + self.q_flow + self.q_inj + self.i_flow + self.vm_m:
            magnitudes[k] = m.val
            sigma[k] = m.sigma
            k += 1

        return magnitudes, sigma


class StateEstimationResults(PowerFlowResults):

    def __init__(self, n, m, n_tr, bus_names, branch_names, transformer_names, bus_types):
        """

        :param n:
        :param m:
        :param n_tr:
        :param bus_names:
        :param branch_names:
        :param transformer_names:
        :param bus_types:
        """
        # initialize the
        PowerFlowResults.__init__(self,
                                  n=n,
                                  m=m,
                                  n_tr=n_tr,
                                  n_hvdc=0,
                                  bus_names=bus_names,
                                  branch_names=branch_names,
                                  transformer_names=transformer_names,
                                  hvdc_names=(),
                                  bus_types=bus_types)


class StateEstimation(DriverTemplate):

    def __init__(self, circuit: MultiCircuit):
        """
        Constructor
        :param circuit: circuit object
        """

        DriverTemplate.__init__(self, grid=circuit)

        self.results = None

    @staticmethod
    def collect_measurements(circuit: MultiCircuit, bus_idx, branch_idx):
        """
        Form the input from the circuit measurements
        :return: nothing, the input object is stored in this class
        """
        se_input = StateEstimationInput()

        # collect the bus measurements
        for i in bus_idx:

            for m in circuit.buses[i].measurements:

                if m.measurement_type == MeasurementType.Pinj:
                    se_input.p_inj_idx.append(i)
                    se_input.p_inj.append(m)

                elif m.measurement_type == MeasurementType.Qinj:
                    se_input.q_inj_idx.append(i)
                    se_input.q_inj.append(m)

                elif m.measurement_type == MeasurementType.Vmag:
                    se_input.vm_m_idx.append(i)
                    se_input.vm_m.append(m)

                else:
                    raise Exception('The bus ' + str(circuit.buses[i]) + ' contains a measurement of type '
                                    + str(m.measurement_type))

        # collect the branch measurements
        branches = circuit.get_branches()
        for i in branch_idx:

            # branch = circuit.branches[i]

            for m in branches[i].measurements:

                if m.measurement_type == MeasurementType.Pflow:
                    se_input.p_flow_idx.append(i)
                    se_input.p_flow.append(m)

                elif m.measurement_type == MeasurementType.Qflow:
                    se_input.q_flow_idx.append(i)
                    se_input.q_flow.append(m)

                elif m.measurement_type == MeasurementType.Iflow:
                    se_input.i_flow_idx.append(i)
                    se_input.i_flow.append(m)

                else:
                    raise Exception('The branch ' + str(branches[i]) + ' contains a measurement of type '
                                    + str(m.measurement_type))

        return se_input

    def run(self):
        """
        Run state estimation
        :return:
        """
        n = len(self.grid.buses)
        m = self.grid.get_branch_number()

        numerical_circuit = compile_snapshot_circuit(self.grid)
        self.results = StateEstimationResults(n=n, m=m,
                                              n_tr=numerical_circuit.ntr,
                                              bus_names=numerical_circuit.bus_names,
                                              branch_names=numerical_circuit.branch_names,
                                              transformer_names=numerical_circuit.transformer_data.tr_names,
                                              bus_types=numerical_circuit.bus_types)
        # self.se_results.initialize(n, m)

        islands = numerical_circuit.split_into_islands()

        self.results.bus_types = numerical_circuit.bus_types

        for island in islands:

            # collect inputs of the island
            se_input = self.collect_measurements(circuit=self.grid,
                                                 bus_idx=island.original_bus_idx,
                                                 branch_idx=island.original_branch_idx)

            # run solver
            report = bs.ConvergenceReport()
            solution = solve_se_lm(Ybus=island.Ybus,
                                   Yf=island.Yf,
                                   Yt=island.Yt,
                                   f=island.F,
                                   t=island.T,
                                   se_input=se_input,
                                   ref=island.vd,
                                   pq=island.pq,
                                   pv=island.pv)

            report.add(method=bs.SolverType.LM,
                       converged=solution.converged,
                       error=solution.norm_f,
                       elapsed=solution.elapsed,
                       iterations=solution.iterations)

            # Compute the branches power and the slack buses power
            Sfb, Stb, If, It, Vbranch, loading, losses, Sbus = power_flow_post_process(calculation_inputs=island,
                                                                                       Sbus=island.Sbus,
                                                                                       V=solution.V,
                                                                                       branch_rates=island.branch_rates,
                                                                                       Yf=None, Yt=None)

            # pack results into a SE results object
            results = StateEstimationResults(n=island.nbus,
                                             m=island.nbr,
                                             n_tr=island.ntr,
                                             bus_names=island.bus_names,
                                             branch_names=island.branch_names,
                                             transformer_names=island.transformer_data.tr_names,
                                             bus_types=island.bus_types)
            results.Sbus = Sbus
            results.Sf = Sfb
            results.voltage = solution.V
            results.losses = losses
            results.loading = loading
            results.convergence_reports.append(report)

            self.results.apply_from_island(results,
                                           island.original_bus_idx,
                                           island.original_branch_idx,
                                           island.original_tr_idx)


if __name__ == '__main__':
    pass